# ----------------------------------------------------
# Copyright The IETF Trust 2018-9, All Rights Reserved
# ----------------------------------------------------

import pycodestyle
import unittest
import os
import shutil
import difflib
import sys
import subprocess
import six
import inspect
import struct
import re
from rfctools_common.parser import XmlRfcParser
from xmldiff.EditItem import EditItem
from xmldiff.zzs2 import distance
from xmldiff.DiffNode import DiffRoot, BuildDiffTree
from xmldiff.DiffNode import ChangeTagMatching, AddParagraphs, SourceFiles
from xmldiff.EditDistance import DoWhiteArray, ComputeEdits

xmldiff_program = "rfc-xmldiff"


def which(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
    return None


class OOO(object):
    def __init__(self):
        self.debug = True


class TestCommandLineOptions(unittest.TestCase):
    """ Run a set of command line checks to make sure they work """
    def test_get_version(self):
        check_process(self, [sys.executable, xmldiff_program, "--version"],
                      "Results/Version.out", "Results/Empty.txt",
                      None, None)

    def test_clear_cache(self):
        if not os.path.exists('Temp'):
            os.mkdir('Temp')
        if not os.path.exists('Temp/cache'):
            os.mkdir('Temp/cache')
        shutil.copy('Tests/cache_saved/reference.RFC.1847.xml',
                    'Temp/cache/reference.RFC.1847.xml')
        check_process(self, [sys.executable, xmldiff_program, "--clear-cache",
                             "--cache=Temp/cache"],
                      None, None,
                      None, None)
        self.assertFalse(os.path.exists('Temp/cache/reference.RFC.1847.xml'))

    def test_help(self):
        check_process(self, [sys.executable, xmldiff_program, "--help"],
                      "Results/Help.out", "Results/Empty.txt", None, None)

    def test_base_default(self):
        if not os.path.exists('Temp'):
            os.mkdir('Temp')
        check_process(self, [sys.executable, xmldiff_program, "--quiet",
                             "-o", "Temp/Single.html", "Tests/Simple.xml", "Tests/SimpleTree.xml"],
                      "Results/Empty.txt", "Results/Empty.txt",
                      "Temp/Base.html", "Results/Base.html")

    def test_single(self):
        if not os.path.exists('Temp'):
            os.mkdir('Temp')
        check_process(self, [sys.executable, xmldiff_program, "-t", "single.html", "--quiet",
                             "-o", "Temp/Single.html", "Tests/Simple.xml", "Tests/SimpleTree.xml"],
                      "Results/Empty.txt", "Results/Empty.txt",
                      "Temp/Single.html", "Results/Single.html")

    def test_base(self):
        if not os.path.exists('Temp'):
            os.mkdir('Temp')
        check_process(self, [sys.executable, xmldiff_program, "-t", "base.html", "--quiet",
                             "-o", "Temp/Base.html", "Tests/Simple.xml", "Tests/SimpleTree.xml"],
                      "Results/Empty.txt", "Results/Empty.txt",
                      "Temp/Base.html", "Results/Base.html")


class Test_Coding(unittest.TestCase):

    def test_pycodestyle_conformance(self):
        """Test that we conform to PEP8."""
        dirParent = os.path.dirname(os.getcwd())

        files = [f for f in os.listdir(os.getcwd()) if f[-3:] == '.py' and f[:8] != 'torture.']

        errors = 0
        pep8style = pycodestyle.StyleGuide(quiet=False, config_file="pycode.cfg")
        result = pep8style.check_files(files)

        errors += result.total_errors

        files = [os.path.join(dirParent, f) for f in os.listdir(dirParent) if f[-3:] == '.py']
        pep8style = pycodestyle.StyleGuide(quiet=False, config_file="pycode.cfg")
        result = pep8style.check_files(files)
        errors += result.total_errors

        dirParent = os.path.join(dirParent, "xmldiff")
        files = [os.path.join(dirParent, f) for f in os.listdir(dirParent)
                 if f[-3:] == '.py']
        pep8style = pycodestyle.StyleGuide(quiet=False, config_file="pycode.cfg")
        result = pep8style.check_files(files)
        errors += result.total_errors

        self.assertEqual(errors, 0,
                         "Found code style errors (and warnings).")

    def test_pyflakes_confrmance(self):
        files = [f for f in os.listdir(os.getcwd()) if f[-3:] == '.py' and f[:8] != 'torture.']
        dir = os.path.basename(os.getcwd())
        dirParent = os.path.dirname(os.getcwd())

        files2 = [os.path.join(dirParent, f) for f in os.listdir(dirParent) if f[-3:] == '.py']
        files.extend(files2)

        dir = os.path.join(dirParent, 'xmldiff')
        files2 = [os.path.join(dir, f) for f in os.listdir(dir)
                  if f[-3:] == '.py']
        files.extend(files2)

        files.insert(0, 'pyflakes')

        p = subprocess.Popen(files, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        (stdoutX, stderrX) = p.communicate()
        ret = p.wait()
        if ret > 0:
            if six.PY3:
                stdoutX = stdoutX.decode('utf-8')
                stderrX = stderrX.decode('utf-8')
            print(stdoutX)
            print(stderrX)
            self.assertEqual(ret, 0)

    def test_copyright(self):
        dir = os.path.basename(os.getcwd())
        dirParent = os.path.dirname(os.getcwd())
        if dir != 'Test':
            return
        files = [f for f in os.listdir(os.getcwd()) if f[-3:] == '.py']

        copyright_year_re = r"(?i)Copyright The IETF Trust 201\d-%s, All Rights Reserved" % (9)

        passed = True

        for name in files:
            with open(name) as file:
                try:
                    chunk = file.read(4000)
                except UnicodeDecodeError:
                    print("Error reading file %s" % (name))
                    passed = False
                    continue
                if not re.search(copyright_year_re, chunk):
                    print("No copyright in file %s" % (name))
                    passed = False
                    continue

        files = [f for f in os.listdir(dirParent) if f[-3:] == '.py']

        for name in files:
            with open(os.path.join("..", name)) as file:
                try:
                    chunk = file.read(4000)
                except UnicodeDecodeError:
                    print("Error reading file %s" % (name))
                    passed = False
                    continue
                if not re.search(copyright_year_re, chunk):
                    print("No copyright in file %s" % (name))
                    passed = False
                    continue

        path2 = os.path.join(dirParent, "xmldiff")
        files = [f for f in os.listdir(path2) if f[-3:] == '.py']

        for name in files:
            with open(os.path.join(path2, name)) as file:
                try:
                    chunk = file.read(4000)
                except UnicodeDecodeError:
                    print("Error reading file %s" % (name))
                    passed = False
                    continue
                if not re.search(copyright_year_re, chunk):
                    print("No copyright in file %s" % (name))
                    passed = False
                    continue

        self.assertTrue(passed)


class TestDistanceMethods(unittest.TestCase):
    """ Set of simple functions to test the distance functions """
    def test_SingleFile(self):
        """ Compare file with self """
        DistanceTest(self, "Tests/SimpleTree.xml", "Tests/SimpleTree.xml",
                     "Results/SimpleDistance.txt", "Results/SimpleDistance.html", False)

    def test_Add1(self):
        """ Insert one inline node """
        DistanceTest(self, "Tests/SimpleTree.xml", "Tests/Simple-AddNode.xml",
                     "Results/Simple-Add1.txt", "Results/Simple-Add1.html", False)

    def test_Remove1(self):
        """ Remove one inline node """
        DistanceTest(self, "Tests/Simple-AddNode.xml", "Tests/SimpleTree.xml",
                     "Results/Simple-Remove1.txt", "Results/Simple-Remove1.html", False)

    def test_Add2(self):
        """ Insert two sibling nodes """
        DistanceTest(self, "Tests/SimpleTree.xml", "Tests/Simple-Add2.xml",
                     "Results/Simple-Add2.txt", "Results/Simple-Add2.html", False)

    def test_Remove2(self):
        """ Remove two sibling node """
        DistanceTest(self, "Tests/Simple-Add2.xml", "Tests/SimpleTree.xml",
                     "Results/Simple-Remove2.txt", "Results/Simple-Remove2.html", False)

    def test_Add3(self):
        """ Insert leaf nodes """
        DistanceTest(self, "Tests/Simple.xml", "Tests/SimpleTree.xml",
                     "Results/Simple-Add3.txt", "Results/Simple-Add3.html", False)

    def test_Remove3(self):
        """ Remove leaf node """
        DistanceTest(self, "Tests/SimpleTree.xml", "Tests/Simple.xml",
                     "Results/Simple-Remove3.txt", "Results/Simple-Remove3.html", False)

    def test_Add4(self):
        """ Insert four sibling nodes """
        DistanceTest(self, "Tests/SimpleTree.xml", "Tests/Simple-Add3.xml",
                     "Results/Simple-Add4.txt", "Results/Simple-Add4.html", False)

    def test_Remove4(self):
        """ Remove four sibling node """
        DistanceTest(self, "Tests/Simple-Add3.xml", "Tests/SimpleTree.xml",
                     "Results/Simple-Remove4.txt", "Results/Simple-Remove4.html", False)

    def test_Add5(self):
        """ Replace text with nodes """
        DistanceTest(self, "Tests/SimpleTree.xml", "Tests/Simple-Add4.xml",
                     "Results/Simple-Add5.txt", "Results/Simple-Add5.html", False)

    def test_Remove5(self):
        """ Replace nodes w/ text """
        DistanceTest(self, "Tests/Simple-Add4.xml", "Tests/SimpleTree.xml",
                     "Results/Simple-Remove5.txt", "Results/Simple-Remove5.html", False)

    def test_AddAttr1(self):
        """ Insert attribute """
        DistanceTest(self, "Tests/SimpleTree.xml", "Tests/AttrTree1.xml",
                     "Results/Simple-AddAttr1.txt", "Results/Simple-AddAttr1.html", False)

    def test_RemoveAddr1(self):
        """ Remove attribute """
        DistanceTest(self, "Tests/AttrTree1.xml", "Tests/SimpleTree.xml",
                     "Results/Simple-RemoveAttr1.txt", "Results/Simple-RemoveAttr1.html", False)

    def test_RenameAttr1(self):
        """ rename attribute """
        DistanceTest(self, "Tests/AttrTree1.xml", "Tests/AttrTree2.xml",
                     "Results/RenameAttr1.txt", "Results/RenameAttr1.html", False)

    def test_AddAttr2(self):
        """ Insert attribute """
        DistanceTest(self, "Tests/SimpleTree.xml", "Tests/AttrTree1.xml",
                     "Results/Simple-AddAttr1.txt", "Results/Simple-AddAttr1.html", False)

    def test_RemoveAddr2(self):
        """ Remove attribute """
        DistanceTest(self, "Tests/AttrTree1.xml", "Tests/AttrTree3.xml",
                     "Results/Simple-RemoveAttr2.txt", "Results/Simple-RemoveAttr2.html", False)

    def test_Insert1(self):
        """ Add a layer to a tree """
        DistanceTest(self, "Tests/Simple-Add3.xml", "Tests/Insert1.xml",
                     "Results/Insert1.txt", "Results/Insert1.html", False)

    def test_Insert2(self):
        """ Add a layer to a tree """
        DistanceTest(self, "Tests/Simple-Add3.xml", "Tests/Insert2.xml",
                     "Results/Insert2.txt", "Results/Insert2.html", False)

    def test_Insert3(self):
        """ Add a layer to a tree """
        DistanceTest(self, "Tests/Simple-Add3.xml", "Tests/Insert3.xml",
                     "Results/Insert3.txt", "Results/Insert3.html", False)

    def test_Insert4(self):
        """ Add a layer to a tree """
        DistanceTest(self, "Tests/Simple-Add3.xml", "Tests/Insert4.xml",
                     "Results/Insert4.txt", "Results/Insert4.html", False)

    def test_Insert5(self):
        """ Add a layer to a tree """
        DistanceTest(self, "Tests/Simple-Add3.xml", "Tests/Insert5.xml",
                     "Results/Insert5.txt", "Results/Insert5.html", False)

    def test_Namespace(self):
        """ Add a layer to a tree """
        DistanceTest(self, "Tests/Namespace1.xml", "Tests/Namespace2.xml",
                     "Results/Namespace.txt", "Results/Namespace.xml", False)

    def test_Table1(self):
        """ Add a layer to a tree """
        from xmldiff.DiffNode import tagMatching

        hold = tagMatching
        ChangeTagMatching(None)
        DistanceTest(self, "Tests/Table1.xml", "Tests/Table2.xml",
                     "Results/Table1.txt", "Results/Table1.html", False)
        ChangeTagMatching(hold)

    def test_Table2(self):
        """ Add a layer to a tree """
        """ If the first diff fails, make sure that table1 finished correctly """
        DistanceTest(self, "Tests/Table1.xml", "Tests/Table2.xml",
                     "Results/Table2.txt", "Results/Table2.html", False)

    def test_Cdata(self):
        """ Deal with space presevation """
        DistanceTest(self, "Tests/Cdata1.xml", "Tests/Cdata2.xml",
                     "Results/Cdata1.txt", "Results/Cdata1.html", True)

    def test_Comment1(self):
        """ Deal with space presevation """
        DistanceTest(self, "Tests/Comment1.xml", "Tests/Comment2.xml",
                     "Results/Comment1.txt", "Results/Comment1.html", True)

    def test_Entity1(self):
        """ Deal with entities not being expanded """
        check_process(self, [sys.executable, xmldiff_program, "--no-resolve-entities",
                             "Tests/Entity1.xml", "Tests/Entity2.xml", "-o", "Temp/Entity.html"],
                      "Results/Empty.txt", "Results/Entity.err",
                      "Temp/Entity.html", "Results/Entity.html")


class TestOverlappedTrees(unittest.TestCase):
    """ Deal with tests for cases where sub-trees will overlap while merging """
    @unittest.skipIf(os.name == 'nt' and struct.calcsize("P") == 4, "Don't run on Python 32-bit")
    def test_Case1_Forward(self):
        DistanceTest(self, "Tests/LOverlap1.xml", "Tests/ROverlap1.xml",
                     "Results/Case1_Forward.txt", "Results/Case1_Forward.xml", True)

    @unittest.skipIf(os.name == 'nt' and struct.calcsize("P") == 4, "Don't run on Python 32-bit")
    def test_Case1_Backward(self):
        DistanceTest(self, "Tests/ROverlap1.xml", "Tests/LOverlap1.xml",
                     "Results/Case1_Backward.txt", "Results/Case1_Backward.xml", True)

    @unittest.skipIf(os.name == 'nt' and struct.calcsize("P") == 4, "Don't run on Python 32-bit")
    def test_Case2_Forward(self):
        DistanceTest(self, "Tests/LOverlap2.xml", "Tests/ROverlap2.xml",
                     "Results/Case2_Forward.txt", "Results/Case2_Forward.xml", True)

    @unittest.skipIf(os.name == 'nt' and struct.calcsize("P") == 4, "Don't run on Python 32-bit")
    def test_Case2_Backward(self):
        DistanceTest(self, "Tests/ROverlap2.xml", "Tests/LOverlap2.xml",
                     "Results/Case2_Backward.txt", "Results/Case2_Backward.xml", True)


class TestStringFormatting(unittest.TestCase):
    def test_case1(self):
        StringAlignTest(self, "Tests/String1a.txt", "Tests/String1b.txt",
                        "Results/String1.txt", None)


def DistanceTest(tester, leftFile, rightFile, diffFile, htmlFile, markParagraphs):
    """ General distance test function """
    options = OOO()
    SourceFiles.Clear()
    left = XmlRfcParser(leftFile, quiet=True, cache_path=None, no_network=True). \
        parse(strip_cdata=False, remove_comments=False)
    left = BuildDiffTree(left.tree, options)
    if markParagraphs:
        left = AddParagraphs(left)
    SourceFiles.LeftDone()
    right = XmlRfcParser(rightFile, quiet=True, cache_path=None, no_network=True). \
        parse(strip_cdata=False, remove_comments=False)
    right = BuildDiffTree(right.tree, options)
    if markParagraphs:
        right = AddParagraphs(right)

    editSet = distance(left, right, DiffRoot.get_children,
                       DiffRoot.InsertCost, DiffRoot.DeleteCost, DiffRoot.UpdateCost)
    with open(diffFile, 'r') as f:
        lines2 = f.readlines()

    d = difflib.Differ()

    # check that the edit set is the same
    lines2 = [line.strip() for line in lines2]
    lines1 = [edit.toString() for edit in editSet]
    result = list(d.compare(lines1, lines2))

    hasError = False
    for ll in result:
        if ll[0:2] == '+ ' or ll[0:2] == '- ':
            hasError = True
            break
    if hasError:
        print("\n".join(result))
        tester.assertFalse(hasError, "edit map differs")

    # check that the HTML output is the same
    left.applyEdits(editSet)
    x = left.ToString()
    x = x.splitlines()
    x = [line.rstrip() for line in x]
    with open(htmlFile, 'rb') as f:
        lines2 = f.read().decode('utf-8').splitlines()
    lines2 = [line.rstrip() for line in lines2]

    result = list(d.compare(x, lines2))

    hasError = False
    for l in result:
        if l[0:2] == '+ ' or l[0:2] == '- ':
            hasError = True
            break
    if hasError:
        print("\n".join(result))
        tester.assertFalse(hasError, "html differs")


def StringAlignTest(tester, leftFile, rightFile, diffFile, htmlFile):
    """ General string comparison function for text alignments """
    if six.PY2:
        with open(leftFile, "rb") as f:
            leftLines = f.read().decode('utf8').replace("\r\n", "\n")
        with open(rightFile, "rb") as f:
            rightLines = f.read().decode('utf8').replace("\r\n", "\n")
        with open(diffFile, "rb") as f:
            opsLines = f.read().decode('utf8').splitlines()
    else:
        with open(leftFile, "r", encoding="utf8") as f:
            leftLines = f.read()
        with open(rightFile, "r", encoding="utf8") as f:
            rightLines = f.read()
        with open(diffFile, "r", encoding="utf8") as f:
            opsLines = f.read().splitlines()

    leftArray = DoWhiteArray(leftLines)
    rightArray = DoWhiteArray(rightLines)

    ops = ComputeEdits(leftArray, rightArray)

    opLines = [u"{0} {1:2d} {2:2d} {3:2d} {4:2d} '{5}' '{6}'".
               format(op[0], op[1], op[2], op[3], op[4],
                      ''.join(leftArray[op[1]:op[2]]).replace('\n', "\\n"),
                      ''.join(rightArray[op[3]:op[4]]).replace('\n', '\\n'))
               for op in ops]

    d = difflib.Differ()
    result = list(d.compare(opLines, opsLines))

    hasError = False
    for l in result:
        if l[0:2] == '+ ' or l[0:2] == '- ':
            hasError = True
            break

    if hasError:
        # print(generatedFile)
        print(u"\n".join(result))

    tester.assertFalse(hasError, "Comparisons failed")


def check_process(tester, args, stdoutFile, errFile, generatedFile, compareFile):
    """
    Execute a subprocess using args as the command line.
    if stdoutFile is not None, compare it to the stdout of the process
    if generatedFile and compareFile are not None, compare them to each other
    """

    if args[1][-4:] == '.exe':
        args = args[1:]
    p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (stdoutX, stderr) = p.communicate()
    p.wait()

    returnValue = True
    if stdoutFile is not None:
        with open(stdoutFile, 'r') as f:
            lines2 = f.readlines()

        if six.PY2:
            lines1 = stdoutX.splitlines(True)
        else:
            lines1 = stdoutX.decode('utf-8').splitlines(True)

        if os.name == 'nt':
            lines2 = [line.replace('Tests/', 'Tests\\').replace('Temp/', 'Temp\\')
                      for line in lines2]
            lines1 = [line.replace('\r', '') for line in lines1]

        d = difflib.Differ()
        result = list(d.compare(lines1, lines2))

        hasError = False
        for l in result:
            if l[0:2] == '+ ' or l[0:2] == '- ':
                hasError = True
                break
        if hasError:
            print("stdout:")
            print("".join(result))
            returnValue = False

    if errFile is not None:
        with open(errFile, 'r') as f:
            lines2 = f.readlines()

        if six.PY2:
            lines1 = stderr.splitlines(True)
        else:
            lines1 = stderr.decode('utf-8').splitlines(True)

        if os.name == 'nt':
            lines2 = [line.replace('Tests/', 'Tests\\').replace('Temp/', 'Temp\\')
                      for line in lines2]
            lines1 = [line.replace('\r', '') for line in lines1]

        d = difflib.Differ()
        result = list(d.compare(lines1, lines2))

        hasError = False
        for l in result:
            if l[0:2] == '+ ' or l[0:2] == '- ':
                hasError = True
                break
        if hasError:
            print("stderr")
            print("".join(result))
            returnValue = False

    if generatedFile is not None:
        template_dir = os.path.abspath(os.path.dirname(inspect.getmodule(EditItem).__file__))
        template_dir = os.path.abspath(os.path.dirname(inspect.getsourcefile(EditItem)))
        if os.name == 'nt':
            template_dir = template_dir.replace('\\', '/')
            template_dir = 'file:///' + template_dir.lower()
        else:
            template_dir = 'file://' + template_dir.lower()

        with open(generatedFile, 'r') as f:
            lines2 = f.readlines()
        lines2 = [line.lower() if 'file://' in line else line for line in lines2]

        with open(compareFile, 'r') as f:
            lines1 = f.readlines()
        lines1 = [line.replace('$TDIR', template_dir) for line in lines1]

        cwd = os.getcwd()
        if os.name == 'nt':
            cwd = cwd.replace('\\', '/')
        lines1 = [line.replace('$$CWD$$', cwd) for line in lines1]

        d = difflib.Differ()
        result = list(d.compare(lines1, lines2))

        hasError = False
        for l in result:
            if l[0:2] == '+ ' or l[0:2] == '- ':
                hasError = True
                break

        if hasError:
            print(generatedFile)
            print("".join(result))
            returnValue = False

    tester.assertTrue(returnValue, "Comparisons failed")


def clear_cache(parser):
    parser.delete_cache()


if __name__ == '__main__':
    if os.environ.get('RFCEDITOR_TEST'):
        xmldiff_program = "../xmldiff/run.py"
    else:
        if os.name == 'nt':
            xmldiff_program += '.exe'
        xmldiff_program = which(xmldiff_program)
        if xmldiff_program is None:
            print("Failed to find the rfc-xmldiff for testing")
            xmldiff_program = "../xmldiff/run.py"

    unittest.main(buffer=True)



#!/usr/bin/env python
# ----------------------------------------------------
# Copyright The IETF Trust 2018-9, All Rights Reserved
# ----------------------------------------------------


import optparse
import os
import six
import sys
from rfctools_common.parser import XmlRfcParser, XmlRfcError, CACHES
from rfctools_common.parser import CachingResolver
from rfctools_common import log
from rfctools_common.log import write_to
from xmldiff.DiffNode import DiffRoot, BuildDiffTree, AddParagraphs
from xmldiff.DiffNode import SourceFiles
import string
from xmldiff.zzs2 import distance
# from xmldiff.zzs import distance, EditItem
from xmldiff.__init__ import __version__

if six.PY2:
    from urlparse import urlparse
else:
    from urllib.parse import urlparse

try:
    import debug
    assert debug
except ImportError:
    pass

try:
    from html import escape  # python 3.x
except ImportError:
    from cgi import escape   # pthyon 2.x


def display_version(self, opt, value, parser):
    print(__version__)
    sys.exit()


def clear_cache(cache_path):
    XmlRfcParser('', cache_path=cache_path).delete_cache()
    sys.exit()


def formatLines(lines, side, fileNumber):
    output = '<div itemprop="text" class="blob-wrapper data type-c">\n'
    output += '<table class="highlight tab-size js-file-line-container" data-tab-size="8">\n'
    output += "<col width='4em'/>\n"

    iLine = 1
    for line in lines:
        output += u'<tr id="{2}{4}_{3}"><td class="blob-num js-line-number" ' \
                  'data-line-number="{0}">{0}</td>' \
                  '<td class="blob-code blob-code-inner js-file-line">{1}</td></tr>\n'. \
                  format(iLine, line, side, fileNumber, iLine-1)
        iLine += 1
    output += "</table>\n</div>\n"

    return output


def main():
    """ Main function for xmldiff """

    formatter = optparse.IndentedHelpFormatter(max_help_position=40)
    optionparser = optparse.OptionParser(usage='xmldiff LEFT RIGHT [OPTIONS] '
                                         '...\nExample: rfc-xmldiff '
                                         'draft1.xml draft2.xml',
                                         formatter=formatter)

    value_options = optparse.OptionGroup(optionparser, 'Other Options')
    value_options.add_option('-o', '--out', dest='output_filename', metavar='FILE',
                             help='specify an explicit output filename',
                             default=None)
    value_options.add_option('--debug', action="store_true",
                             help='Show debugging output')
    value_options.add_option('--raw', action="store_true",
                             help='Diff using the raw tree')
    value_options.add_option('-t', '--template', dest='template', metavar='FILE',
                             help='specify the HTML template filename',
                             default='base.html')
    value_options.add_option('-V', '--version', action='callback', callback=display_version,
                             help='display the version number and exit')
    value_options.add_option('-C', '--clear-cache', action='store_true', dest='clear_cache',
                             default=False, help='purge the cache and exit')
    value_options.add_option('-c', '--cache', dest='cache',
                             help='specify a primary cache directory to write to; '
                             'default: try [ %s ]' % ', '.join(CACHES))
    value_options.add_option('-q', '--quiet', action='store_true',
                             help='dont print anything')
    value_options.add_option('-v', '--verbose', action='store_true',
                             help='print extra information')
    value_options.add_option('--no-resolve-entities', dest='noEntity',
                             action="store_true",
                             help="Don't resolve entities in the XML")
    value_options.add_option('-N', '--no-network', action='store_true', default=False,
                             help='don\'t use the network to resolve references')
    value_options.add_option('-X', '--no-xinclude', action='store_true', dest='no_xinclude',
                             help='don\'t resolve any xi:include elements')
    value_options.add_option('-D', '--no-defaults', action='store_false', default=True,
                             help="don't add default attributes")

    optionparser.add_option_group(value_options)

    # --- Parse and validate arguments ----------------------------

    (options, args) = optionparser.parse_args()

    if options.clear_cache:
        clear_cache(options.cache)

    if len(args) < 1:
        optionparser.print_help()
        sys.exit(2)

    # Setup warnings module
    # rfclint.log.warn_error = options.warn_error and True or False
    log.quiet = options.quiet and True or False
    log.verbose = options.verbose

    # Load the left file
    leftSource = args[0]
    if not os.path.exists(leftSource):
        sys.exit('No such file: ' + leftSource)

    log.note("Parse input files")
    parser = XmlRfcParser(leftSource, verbose=log.verbose,
                          quiet=log.quiet, no_network=options.no_network,
                          no_xinclude=options.no_xinclude,
                          resolve_entities=not options.noEntity,
                          attribute_defaults=options.no_defaults)
    try:
        ll = parser.parse(remove_pis=False, strip_cdata=False, remove_comments=False).tree
        leftXml = BuildDiffTree(ll, options)
        if not options.raw:
            leftXml = AddParagraphs(leftXml)
        leftFile_base = os.path.basename(leftSource)
        SourceFiles.LeftDone()
    except XmlRfcError as e:
        log.exception('Unable to parse the XML document: ' + leftSource, e)
        sys.exit(1)

    rightSource = args[1]
    if not os.path.exists(rightSource):
        sys.exit('No such file: ' + rightSource)

    parser = XmlRfcParser(rightSource, verbose=log.verbose,
                          quiet=log.quiet, no_network=options.no_network,
                          no_xinclude=options.no_xinclude,
                          resolve_entities=not options.noEntity,
                          attribute_defaults=options.no_defaults)
    try:
        rightXml = parser.parse(remove_pis=False, strip_cdata=False, remove_comments=False)
        rightXml = BuildDiffTree(rightXml.tree, options)
        if not options.raw:
            rightXml = AddParagraphs(rightXml)
        rightFile_base = os.path.basename(rightSource)
    except XmlRfcError as e:
        log.exception('Unable to parse the XML document: ' + rightSource, e)
        sys.exit(1)

    log.note("Read files for source display")
    cache = CachingResolver(library_dirs=[])

    leftSources = ""
    leftSourceNames = ""
    for i in range(len(SourceFiles.leftFiles)):
        file = SourceFiles.leftFiles[i]
        if file[:5] == 'file:':
            file = urlparse(file)
            file = file[2]
            if file[2] == ':':
                file = file[1:]
        else:
            file = cache.getReferenceRequest(file)[0]

        if six.PY2:
            with open(file, "rb") as f:
                leftLines = f.read()
                leftLines = leftLines.decode('utf8').splitlines(1)
        else:
            with open(file, "rU", encoding="utf8") as f:
                leftLines = f.readlines()

        leftSources += u'<div id="L_File{0}" class="tabcontent">\n'.format(i)
        leftLines = [escape(x).replace(' ', '&nbsp;').replace('"', '&quot;') for x in leftLines]
        leftSources += formatLines(leftLines, 'L', i)
        leftSources += u'</div>\n'

        leftSourceNames += u'<option label="{0}" value="L_File{1}">{2}</option>\n'. \
                           format(file, i, file)

    rightSources = ""
    rightSourceNames = ""

    for i in range(len(SourceFiles.rightFiles)):
        file = SourceFiles.rightFiles[i]
        if file[:5] == 'file:':
            file = urlparse(file)
            file = file[2]
            if file[2] == ':':
                file = file[1:]
        else:
            file = cache.getReferenceRequest(file)[0]

        if six.PY2:
            with open(file, "rb") as f:
                rightLines = f.read().decode('utf8').splitlines(1)
        else:
            with open(file, "rU", encoding="utf8") as f:
                rightLines = f.readlines()

        rightSources += u'<div id="R_File{0}" class="tabcontent">\n'.format(i)
        rightLines = [escape(x).replace(' ', '&nbsp;').replace('"', '&quot;') for x in rightLines]
        rightSources += formatLines(rightLines, 'R', i)
        rightSources += u'</div>\n'

        rightSourceNames += '<option label="{0}" value="R_File{1}">{2}</option>\n'. \
                            format(file, i, file)

    log.note("Start computing tree edit distance")
    editSet = distance(leftXml, rightXml, DiffRoot.get_children, DiffRoot.InsertCost,
                       DiffRoot.DeleteCost, DiffRoot.UpdateCost)

    if options.debug:
        print("edit count = " + str(len(editSet)))
        for edit in editSet:
            print(edit.toString())

    log.note("Apply copmuted tree edits")
    if len(editSet) == 0:
        log.info("Files are identical")

    leftXml.applyEdits(editSet)

    log.note("Setup to write html file")
    templates_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'Templates')
    log.note("   template directory = " + templates_dir)

    if six.PY2:
        with open(os.path.join(templates_dir, "resize.js"), "rb") as f:
            allScript = f.read()
    else:
        with open(os.path.join(templates_dir, "resize.js"), "rU", encoding="utf8") as f:
            allScript = f.read()

    template_file = options.template
    if not os.path.exists(options.template):
        template_file = os.path.join(templates_dir, options.template)
        if not os.path.exists(template_file):
            sys.exit('No template file: ' + template_file)

    log.note('   template source file: ' + template_file)
    with open(template_file, 'rb') as file:
        html_template = string.Template(file.read().decode('utf8'))

    rightLines = [x.replace(' ', '&nbsp;') for x in rightLines]

    buffers = {}
    buffers['leftFile'] = leftSources
    buffers['rightFile'] = rightSources
    buffers['body'] = leftXml.ToString()

    subs = {
        'background': '',
        # HTML-escaped values
        'title': 'rfc-xmldiff {0} {1}'.format(leftFile_base, rightFile_base),
        'body': ''.join(buffers['body']),
        'leftFile': buffers['leftFile'],
        'leftSourceNames': leftSourceNames,
        'rightFile': buffers['rightFile'],
        'rightSourceNames': rightSourceNames,
        'allScript': allScript
        }
    output = html_template.substitute(subs)

    if options.output_filename is None:
        write_to(sys.stdout, output)
    else:
        log.note('Write out html file: ' + options.output_filename)
        file = open(options.output_filename, "wb")
        file.write(output.encode('utf8'))
        file.close()


if __name__ == '__main__':
    main()



# ----------------------------------------------------
# Copyright The IETF Trust 2018-9, All Rights Reserved
# ----------------------------------------------------

import lxml.etree
from rfctools_common import log
import math
import sys
import difflib
import six
from lxml.html import builder as E
from xmldiff.EditItem import EditItem
from xmldiff.EditDistance import ComputeEdits, DoWhiteArray
# from xmldiff.zzs import EditItem

try:
    import debug
    assert debug
except ImportError:
    pass


TextElements = [
    "bcp14", "br", "cref", "em", "eref", "iref", "strong", "relref", "sub",
    "sup", "tt", "xref", "vspace"
]

V2NonTextElements = [
    "list", "figure"
]

tagMatching = {
    "c": {"td"},
    "dd": {"t"},
    "dl": {"list"},
    "li": {"t"},
    "list": {"dl", "ol", "ul"},
    "ol": {"list"},
    "t": {"dd", "li"},
    "table": {"texttable"},
    "td": {"c"},
    "texttable": {"table"},
    "th": {"ttcol"},
    "ttcol": {"th"},
    "ul": {"list"},
}

PreserveSpace = ['artwork', 'sourcecode']

nsKeys = {
    "http://www.w3.org/XML/1998/namespace": "xml",
    "http://www.w3.org/2001/XInclude": "xi",
    "http://www.w3.org/2003/XInclude": "xi",
    "http://www.w3.org/2003/Xinclude": "xi",
    "http://www.w3.org/2000/svg": "svg"
}
nsKeysIndex = 0

diffCount = 0
nbsp = six.unichr(0xa0)


def ChangeTagMatching(newMatching):
    global tagMatching

    tagMatching = newMatching


class SourceFileSet(object):
    def __init__(self):
        self.leftFiles = []
        self.rightFiles = []
        self.side = True  # Left

    def AddSourceFile(self, fileName):
        if self.side:
            if fileName not in self.leftFiles:
                self.leftFiles.append(fileName)
        else:
            if fileName not in self.rightFiles:
                self.rightFiles.append(fileName)

    def LeftDone(self):
        self.side = False

    def LineFormat(self, xmlNode, fRight):
        if hasattr(xmlNode, "base"):
            if fRight:
                fileNo = self.rightFiles.index(xmlNode.base)
                return "R{0}_{1}".format(xmlNode.sourceline, fileNo)
            else:
                fileNo = self.leftFiles.index(xmlNode.base)
                return "L{0}_{1}".format(xmlNode.sourceline, fileNo)
        else:
            return ("R{0}_0" if fRight else "L{0}_0").format(xmlNode.sourceline)

    def Clear(self):
        self.leftFiles = []
        self.rightFiles = []
        self.side = True  # Left


SourceFiles = SourceFileSet()


def BuildDiffTree(xmlNode, options):
    """ Build the Diff tree from the xml tree """

    global diffCount, SourceFiles

    if not isinstance(xmlNode, lxml.etree._ElementTree):
        sys.exit(2, "bad parameter")

    #  We have two problems here - the first is that lxml does
    #  not bother keeping the root elements as separate things
    #  so life is a pain

    diffCount = 0
    root = DiffDocument(xmlNode)

    # Grab nodes before the root
    element = xmlNode.getroot().getprevious()
    while element is not None:
        if element.tag is lxml.etree.PI:
            root.children.insert(0, DiffPI(element, root))

        element = element.getprevious()

    # Now process the root element itself
    root.children.append(root.createNode(xmlNode.getroot(), root))
    root.children[-1].root = True

    # Now process any elements after to root element

    element = xmlNode.getroot().getnext()
    while element is not None:
        if element.tag is lxml.etree.PI:
            root.children.append(DiffPI(element, root))

        element = element.getnext()

    if options.debug:
        print("Number of diff nodes is {0}".format(diffCount))

    return root

# Control how elements are converted from a tree to a pargaraph.
# The values here mean:
# 1 - This is always just converted to a paragraph.
# 2 - These are v3 elements which can be either a paragraph or
#     a tree (i.e. have other items inside which are paragraphs)
# 3 - These are v2 elements which may not be paragraphs
# 4 - These are v3 element which are paragraphs, but v2 elements
#     which may not be paragraphs.
#
# We ignore those elements which can only simple text and do not
# treat them as paragarphs


ParagraphMarkers = {
    'annotation': 1,
    'artwork': 2,  # need to deal with perserveSpace
    'blockquote': 2,
    'c': 3,
    'dd': 2,
    'dt': 1,
    'li': 2,
    'refcontent': 1,
    'sourcecode': 1,  # need to deal with perserveSpace
    't': 4,
    'td': 2,
    'th': 2,
    'ttcol': 3,
    '{http://relaxng.org/ns/structure/1.0}desc': 1,
    '{http://relaxng.org/ns/structure/1.0}svgTitle': 1,
    '{http://relaxng.org/ns/structure/1.0}a': 1,
    '{http://relaxng.org/ns/structure/1.0}text': 1,
    '{http://relaxng.org/ns/structure/1.0}textArea': 1,
    '{http://relaxng.org/ns/structure/1.0}tspan': 1
    }


def AddParagraphs(root):
    if not isinstance(root, DiffElement):
        for c in root.children:
            AddParagraphs(c)
        return root
    if root.xml.tag in ParagraphMarkers:
        if ParagraphMarkers[root.xml.tag] == 1 or \
           ParagraphMarkers[root.xml.tag] == 3:
            p = DiffParagraph(root.xml, root)
            p.children = root.children
            root.children = [p]
            return root
        elif ParagraphMarkers[root.xml.tag] == 2:
            newChildren = []
            p = DiffParagraph(root.xml, root)
            for child in root.children:
                if isinstance(child, DiffComment) and False:
                    if len(p.children) > 0:
                        newChildren.append(p)
                        p = DiffParagraph(root.xml, root)
                    newChildren.append(child)
                elif not isinstance(child, DiffElement):
                    p.children.append(child)
                elif child.xml.tag in TextElements:
                    p.children.append(child)
                else:
                    if len(p.children) > 0:
                        newChildren.append(p)
                        p = DiffParagraph(root.xml, root)
                    newChildren.append(child)
            if len(p.children) > 0:
                newChildren.append(p)
            root.children = newChildren
        elif ParagraphMarkers[root.xml.tag] == 4:
            newChildren = []
            p = DiffParagraph(root.xml, root)
            for child in root.children:
                if isinstance(child, DiffComment) and False:
                    if len(p.children) > 0:
                        newChildren.append(p)
                        p = DiffParagraph(root.xml, root)
                    newChildren.append(child)
                elif not isinstance(child, DiffElement):
                    p.children.append(child)
                elif child.xml.tag not in V2NonTextElements:
                    p.children.append(child)
                else:
                    if len(p.children) > 0:
                        newChildren.append(p)
                        p = DiffParagraph(root.xml, root)
                    newChildren.append(child)
            if len(p.children) > 0:
                newChildren.append(p)
            root.children = newChildren

    for c in root.children:
        AddParagraphs(c)
    return root


def DecorateSourceFile(diffRoot, sourceLines):
    diffRoot.decorateSource(sourceLines)


class DiffRoot(object):
    """ Root of the diff objects. """

    def __init__(self, xmlNode, parent):
        global diffCount

        self.xml = xmlNode
        self.children = []
        self.deleted = False
        self.inserted = False
        self.matchNode = None
        self.parent = parent
        diffCount += 1
        self.index = diffCount
        self.deleteTree = False
        self.preserve = False
        self.root = False

        if hasattr(xmlNode, "base"):
            baseFile = xmlNode.base
            SourceFiles.AddSourceFile(baseFile)

    def ToString(self):
        node = E.DIV()
        node.text = "Need to override something"
        return node

    def toText(self):
        return lxml.etree.tostring(self.xml)  # with_tail=False

    def ToHtml(self, parent):
        node = E.LI()
        node.text = "Need to override something"
        parent.append(node)

    @staticmethod
    def get_children(n):
        return n.children

    @staticmethod
    def InsertCost(left):
        item = EditItem()
        item.setOperation(EditItem.OP_INSERT, None, left)
        if isinstance(left, DiffElement) or isinstance(left, DiffComment) or \
           isinstance(left, DiffText) or isinstance(left, DiffParagraph):
            item.cost = 10
        else:
            item.cost = 1
        return item

    @staticmethod
    def DeleteCost(left):
        item = EditItem()
        item.setOperation(EditItem.OP_DELETE, left, None)
        item.cost = left.deleteCost()
        return item

    def deleteCost(self):
        """ Compute the deletion cost of a node
            The deletion cost for an element needs to be higher
            that that for text so that doing a rename on texts
            is the preferred operation
        """
        if isinstance(self, DiffElement) or isinstance(self, DiffComment) or \
           isinstance(self, DiffText) or isinstance(self, DiffParagraph):
            cost = 10
        else:
            cost = 1
        return cost

    @staticmethod
    def UpdateCost(left, right):
        #  If the types are not the same, then the cost is extremely high
        if type(left) is not type(right):
            item = EditItem()
            item.setOperation(EditItem.OP_RENAME, left, right)
            item.cost = 100000
            return item
        return left.updateCost(right)

    def updateCost(self, right):
        return 100

    def _serialize(self, element):
        if sys.version > '3':
            return lxml.html.tostring(element, pretty_print=True, method='html',
                                      encoding='ascii').decode('ascii')
        else:
            return lxml.html.tostring(element, pretty_print=True, method='html',
                                      encoding='ascii').decode('ascii')

    def markInsertTrees(self):
        insertTree = True
        for child in self.children:
            insertTree &= child.markInsertTrees()
        insertTree &= (self.matchNode is None)
        self.insertTree = insertTree
        return insertTree

    def markDeleteTrees(self):
        deleteTree = True
        for child in self.children:
            deleteTree &= child.markDeleteTrees()
        deleteTree &= (self.matchNode is None)
        self.deleteTree = deleteTree
        return deleteTree

    def createNode(self, xml, parent):
        if xml.tag is lxml.etree.PI:
            return DiffPI(xml, parent)
        if xml.tag is lxml.etree.Comment:
            return DiffComment(xml, parent)
        if xml.tag is lxml.etree.Entity:
            return DiffEntity(xml, parent)
        return DiffElement(xml, parent)

    def getParents(self, includeSelf=False):
        retValue = []
        if includeSelf:
            retValue.append(self)
        c = self.parent
        while c is not None:
            retValue.append(c)
            c = c.parent
        return retValue

    #  Does this node preceed the otherNode in
    #  the tree?
    def preceeds(self, otherNode):
        myParents = self.getParents(True)[::-1]
        otherParents = otherNode.getParents(True)[::-1]
        count = min(len(myParents), len(otherParents))
        left = None
        parent = None
        right = None

        for i in range(count):
            if myParents[i] != otherParents[i]:
                left = myParents[i]
                right = otherParents[i]
                parent = myParents[i-1]
                break

        if left is None:
            if len(myParents) > len(otherParents):
                return False
            elif len(myParents) == len(otherParents):
                return False
            else:
                right = otherParents[count-1]
                parent = myParents[-1]
                # should this return True?
            return False
        else:
            iLeft = -1
            iRight = -1
            for i in range(len(parent.children)):
                if right == parent.children[i]:
                    iRight = i
                if left == parent.children[i]:
                    iLeft = i
            return iLeft <= iRight

    def getPredecessor(self, searchNode):
        prevSib = None
        for child in self.children:
            if searchNode.isMatchNode(child):
                return prevSib
            prevSib = child
        return None

    def getSuccessor(self, searchNode):
        found = False
        for child in self.children:
            if found:
                return child
            if searchNode.isMatchNode(child):
                found = True

        return None

    def isMatchNode(self, other):
        if self == other:
            return True
        if other.inserted:
            ancestorList = other.getParents()
            return self in ancestorList
        if self.deleted:
            if len(self.children) != 1:
                return False
            return self.children[0].isMatchNode(other)
        if other.deleted:
            if len(other.children) != 1:
                return False
            return self.isMatchNode(other.children[0])
        return False

    def insertAfter(self, siblingNode, newNode):
        """ Insert newNode after siblingNode """
        i = 0
        for child in siblingNode.parent.children:
            if siblingNode.isMatchNode(child):
                break
            i += 1
        if i == len(siblingNode.parent.children):
            log.error("ICE: insertAfter failed to find node.")
            siblingNode.parent.children.append(newNode)
        else:
            siblingNode.parent.children.insert(i+1, newNode)
        newNode.parent = siblingNode.parent

    def insertBefore(self, siblingNode, newNode):
        i = 0
        for child in siblingNode.parent.children:
            if siblingNode.isMatchNode(child):
                break
            i += 1
        if i == len(siblingNode.parent.children):
            # assert false
            log.error("ICE: insertBefore failed to find node.")
            siblingNode.parent.children.append(newNode)
        else:
            siblingNode.parent.children.insert(i, newNode)
        newNode.parent = siblingNode.parent

    def fixPreserveSpace(self, node, text):
        if self.preserve:
            text = text.splitlines(1)
            n = None
            children = list(node)
            if len(children) > 0:
                n = children[-1]
            for line in text:
                if n is None:
                    if node.text is None:
                        node.text = line.replace(' ', nbsp)
                    else:
                        node.text += line.replace(' ', nbsp)
                else:
                    if n.tail is None:
                        n.tail = line.replace(' ', nbsp)
                    else:
                        n.tail += line.replace(' ', nbsp)
                if line[-1:] == '\n':
                    n = E.BR()
                    node.append(n)
        else:
            if node.text is None:
                node.text = text
            else:
                node.text += text

    def fixPreserveNL(self, node, text):
        if self.preserve or True:
            text = text.splitlines(1)
            n = None
            children = list(node)
            if len(children) > 0:
                n = children[-1]
            for line in text:
                if n is None:
                    if node.text is None:
                        node.text = line
                    else:
                        node.text += line
                else:
                    if n.tail is None:
                        n.tail = line
                    else:
                        n.tail += line
                if line[-1:] == '\n':
                    n = E.BR()
                    node.append(n)
        else:
            if node.text is None:
                node.text = text
            else:
                node.text += text

    def diffTextToHtml(self, leftText, rightText, node):
        if self.preserve:
            n = E.SPAN()
            n.attrib['class'] = 'artwork'
            node.append(n)
            node = n
        if not rightText:
            if not leftText:
                return
            n = E.SPAN()
            n.attrib["class"] = 'left'
            self.fixPreserveSpace(n, leftText)
            node.append(n)
        elif not leftText:
            n = E.SPAN()
            n.attrib['class'] = 'right'
            self.fixPreserveSpace(n, rightText)
            node.append(n)
        else:
            leftArray = self.doWhiteArray(leftText)
            rightArray = self.doWhiteArray(rightText)
            result = list(ComputeEdits(leftArray, rightArray))

            for op, i1, i2, j1, j2 in result:
                if op == 'equal':
                    if i1 == i2 and j1 == j2:
                        continue
                    bothNode = E.SPAN()
                    self.fixPreserveNL(bothNode, (''.join(leftArray[i1:i2])))
                    node.append(bothNode)
                elif op == 'remove':
                    leftNode = E.SPAN()
                    leftNode.attrib['class'] = 'left'
                    self.fixPreserveNL(leftNode, (''.join(leftArray[i1:i2])+u' '))
                    node.append(leftNode)
                elif op == 'insert':
                    rightNode = E.SPAN()
                    rightNode.attrib['class'] = 'right'
                    rightNode.text = ''
                    self.fixPreserveNL(rightNode, (''.join(rightArray[j1:j2])+u' '))
                    node.append(rightNode)
                else:
                    n = E.SPAN()
                    n.attrib['class'] = 'error'
                    self.fixPreserveNL(n, ''.join(leftArray[i1:i2])+u"*" +
                                       (''.join(rightArray[j1:j2])+u' '))
                    node.append(n)

    def doWhiteArray(self, text):
        return DoWhiteArray(text)

    def untangleLeft(self):
        for child in self.children:
            child.parent = self
            child.untangleLeft()
        self.matchNode = None
        self.deleted = True
        if self.inserted:
            newChildren = []
            for child in self.parent.children:
                if child == self:
                    newChildren = newChildren + self.children
                else:
                    newChildren.append(child)
            self.parent.children = newChildren
            self.parent = None
            self.children = None

    def untangleRight(self):
        for child in self.children:
            child.untangleRight()
        self.matchNode = None
        self.insertTree = True

    def untangle(self):
        if self.matchNode:
            self.matchNode.untangleRight()
        self.untangleLeft()


class DiffDocument(DiffRoot):
    """ Represent the XML document.  We want to have a common
        node type that is always at the root for comparisions
    """

    def __init__(self, xmlNode):
        DiffRoot.__init__(self, xmlNode, None)

    def ToString(self):
        result = E.DIV()
        result.attrib['class'] = 'center'

        ul = E.UL()
        result.append(ul)
        ul.attrib['class'] = 'mktree'
        ul.attrib['id'] = 'diffRoot'

        #  Insert xml declarations

        n = E.LI()
        n.attrib['whereLeft'] = "L0_0"
        n.attrib['whereRight'] = "R0_0"
        leftText = '<?xml version="{0}" encoding="{1}"?>'.format(self.xml.docinfo.xml_version,
                                                                 self.xml.docinfo.encoding)
        rightText = '<?xml version="{0}" encoding="{1}"?>'. \
            format(self.matchNode.xml.docinfo.xml_version,
                   self.matchNode.xml.docinfo.encoding)
        self.diffTextToHtml(leftText, rightText, n)
        ul.append(n)

        # Insert DTD declaration if one exists
        n = E.LI()
        if self.xml.docinfo.doctype or self.matchNode.xml.docinfo.doctype:
            self.diffTextToHtml(self.xml.docinfo.doctype, self.matchNode.xml.docinfo.doctype, n)
            ul.append(n)

        # now put all of the children into HTML

        for child in self.children:
            child.ToHtml(ul)

        return self._serialize(result)

    def updateCost(self, right):
        return 0

    def decorateSource(self, sourceLines):
        for child in self.children:
            child.decorateSource(sourceLines)

    def applyEdits(self, editList):
        newEdits = []
        for edit in editList:
            if edit.operation == EditItem.OP_DELETE:
                edit.left.deleted = True
            elif edit.operation == EditItem.OP_MATCH:
                edit.left.matchNode = edit.right
                edit.right.matchNode = edit.left
            elif edit.operation == EditItem.OP_RENAME:
                edit.left.matchNode = edit.right
                edit.right.matchNode = edit.left
            else:
                newEdits.append(edit)

        self.matchNode.markInsertTrees()
        self.markDeleteTrees()

        while True:
            editList = newEdits
            newEdits = []

            for edit in editList:
                # Already processed
                if edit.right.matchNode is not None:
                    continue

                # if isinstance(edit.right, DiffElement) and \
                #    (edit.right.xml.tag == 'back' or edit.right.xml.tag == 'middle'):
                #     thisIsABreak = 9

                if edit.right.insertTree:
                    matchingParent = edit.right.parent.matchNode
                    if matchingParent is None:
                        # we don't know where it goes yet
                        newEdits.append(edit)
                        continue

                    # If a node has no children, then we can add it as a child
                    if edit.right.parent.matchNode.children is None or \
                       len(edit.right.parent.matchNode.children) == 0:
                        if edit.right.parent.matchNode.children is None:
                            edit.right.parent.matchNode.children = []
                        edit.right.parent.matchNode.children.append(
                            edit.right.cloneTree(edit.right.parent.matchNode))
                        continue

                    # If we have a matched preceeding node, put it after that one
                    sibling = edit.right.parent.getPredecessor(edit.right)
                    if sibling is not None:
                        if sibling.matchNode is not None:
                            newNode2 = edit.right.cloneTree(matchingParent)
                            matchingParent.insertAfter(sibling.matchNode, newNode2)
                            continue

                    # If we have a matching successor node, put it before that one
                    sibling = edit.right.parent.getSuccessor(edit.right)
                    if sibling is not None:
                        if sibling.matchNode is not None:
                            newNode2 = edit.right.cloneTree(matchingParent)
                            matchingParent.insertBefore(sibling.matchNode, newNode2)
                            continue

                    # If all of the left children are deleted and a new right is added.
                    allDeleted = True
                    for child in matchingParent.children:
                        if not child.deleteTree:
                            allDeleted = False
                            break

                    if allDeleted:
                        newNode2 = edit.right.cloneTree(matchingParent)
                        matchingParent.children.append(newNode2)
                        continue

                    newEdits.append(edit)
                    continue

                # Nodes which have undealt with children are deferred
                f = True
                for child in edit.right.children:
                    if child.matchNode is None and not child.insertTree:
                        f = False
                        break
                if not f:
                    newEdits.append(edit)
                    continue

                # Get the list of children that we need to match
                matchList = []
                for child in edit.right.children:
                    if not child.inserted and not child.insertTree:
                        matchList.append(child.matchNode)
                if len(matchList) == 0:
                    newEdits.append(edit)
                    continue

                # Build the list of all common ancestors of these nodes
                commonParents = None

                for child in matchList:
                    ancestorList = child.getParents()[::-1]
                    if commonParents is None:
                        commonParents = ancestorList
                    else:
                        if len(ancestorList) < len(commonParents):
                            commonParents = commonParents[:len(ancestorList)]
                        for i in range(0, min(len(ancestorList), len(commonParents))):
                            if ancestorList[i] != commonParents[i]:
                                if i == 0:
                                    pass
                                else:
                                    commonParents = commonParents[:i]
                                break
                if len(commonParents) == 0:
                    # we have no common parents so this is a mess.
                    newEdits.append(edit)
                    continue

                matchParent = commonParents[-1]

                # If we have a preceeding sibling, we need to make sure that we are
                # going to try and insert this after that node.

                sibling = edit.right.parent.getPredecessor(edit.right)

                if sibling and sibling.matchNode:
                    ancestorList = sibling.matchNode.parent.getParents()
                    if matchParent in ancestorList:
                        aList = []
                        bList = []
                        for child in matchList:
                            ancestorList = child.getParents()
                            if sibling.matchNode.parent in ancestorList:
                                aList.append(child)
                            else:
                                bList.append(child)

                        if len(aList) >= len(bList):
                            for child in bList:
                                child.untangle()
                        else:
                            for child in aList:
                                child.untangle()
                        newEdits.append(edit)
                        continue

                    ancestorList = matchParent.getParents(True)
                    if sibling.matchNode in ancestorList:
                        for child in edit.right.children:
                            if child.matchNode:
                                child.matchNode.untangle()
                        edit.right.insertTree = True
                        newEdits.append(edit)
                        continue

                # create the new node
                newNode = edit.right.clone()
                newNode.parent = matchParent
                newNode.inserted = True
                newNode.matchNode = edit.right
                edit.right.matchNode = newNode

                #

                i = 0
                iX = -1
                interums = []
                for child in edit.right.children:
                    if child.insertTree:
                        newNode2 = child.cloneTree(None)
                        newNode.children.append(newNode2)
                        continue
                    while i != len(matchParent.children):
                        if matchParent.children[i].isMatchNode(child.matchNode):
                            if len(interums) != 0:
                                if iX != -1:
                                    for ii in interums:
                                        newNode.children.append(matchParent.children[ii])
                                        matchParent.children[ii] = newNode
                                        del matchParent.children[ii]
                                        i -= 1
                                interums = []
                            newNode.children.append(matchParent.children[i])
                            matchParent.children[i].parent = newNode
                            del matchParent.children[i]
                            if iX == -1:
                                iX = i
                            break
                        else:
                            interums.append(i)
                        i += 1

                if iX == -1:
                    iX = 0
                matchParent.children.insert(iX, newNode)

            if len(editList) == len(newEdits):
                break

        log.note("Number of edits left = " + str(len(newEdits)))
        for edit in newEdits:
            log.note(edit.toString())
        return len(newEdits)


class DiffPI(DiffRoot):
    def __init__(self, xmlNode, parent):
        DiffRoot.__init__(self, xmlNode, parent)

    def ToHtml(self, parent):
        root2 = E.LI()
        parent.append(root2)
        root = E.SPAN()
        root2.append(root)
        if self.inserted:
            root.attrib['class'] = 'right'
            root.attrib["whereRight"] = SourceFiles.LineFormat(self.xml, True)
        elif self.deleted:
            root.attrib['class'] = 'left'
            root.attrib["whereLeft"] = SourceFiles.LineFormat(self.xml, False)
        elif self.matchNode is None:
            root.attrib['class'] = 'error'
        else:
            root.attrib["whereLeft"] = SourceFiles.LineFormat(self.xml, False)
            root.attrib["whereRight"] = SourceFiles.LineFormat(self.matchNode.xml, True)
            if self.xml.target == self.matchNode.xml.target:
                if self.xml.text == self.matchNode.xml.text:
                    pass
                else:
                    root.text = "<?{0} ".format(self.xml.target)
                    s = E.SPAN()
                    s.attrib['class'] = 'left'
                    s.text = self.xml.text
                    root.append(s)
                    s = E.SPAN()
                    s.attrib['class'] = 'right'
                    s.text = self.matchNode.xml.text
                    root.append(s)
                    s.tail = "?>"
                    return
            else:
                root.text = "<?"
                s = E.SPAN()
                s.attrib['class'] = 'left'
                s.text = self.xml.target
                root.append(s)
                s = E.SPAN()
                s.attrib['class'] = 'right'
                s.text = self.matchNode.xml.target
                root.append(s)
                s.tail = ' '
                s = E.SPAN()
                s.attrib['class'] = 'left'
                s.text = self.xml.text
                root.append(s)
                s = E.SPAN()
                s.attrib['class'] = 'right'
                s.text = self.matchNode.xml.text
                root.append(s)
                s.tail = "?>"
                return

        root.text = "<?{0} {1}?>".format(self.xml.target, self.xml.text)

    def cloneTree(self, parent):
        clone = DiffPI(self.xml, parent)
        clone.matchNode = self
        clone.inserted = True
        self.matchNode = clone
        return clone

    def updateCost(self, right):
        if self.xml.target == right.xml.target:
            if self.xml.text == right.xml.text:
                return 0
            else:
                return 50
        return 100


class DiffComment(DiffRoot):
    def __init__(self, xmlNode, parent):
        DiffRoot.__init__(self, xmlNode, parent)
        self.preserve = True

    def ToHtml(self, parent):
        node = E.LI()
        parent.append(node)
        myLine = "<--" + self.xml.text.replace(' ', nbsp) + "-->"
        while myLine[0] == '\n':
            myLine = myLine[1:]
        while myLine[-1] == '\n':
            myLine = myLine[:-1]
        if self.inserted:
            n = E.SPAN()
            n.attrib['class'] = 'artwork right'
            node.attrib["whereRight"] = SourceFiles.LineFormat(self.xml, True)
            self.fixPreserveSpace(n, myLine)
            node.append(n)
        elif self.deleted:
            n = E.SPAN()
            n.attrib['class'] = 'artwork left'
            node.attrib["whereLeft"] = SourceFiles.LineFormat(self.xml, False)
            self.fixPreserveSpace(n, myLine)
            node.append(n)
        elif self.matchNode is None:
            n = E.SPAN()
            n.attrib['class'] = 'artwork error'
            self.fixPreserveSpace(n, myLine)
            node.append(n)
        else:
            node.attrib["whereLeft"] = SourceFiles.LineFormat(self.xml, False)
            node.attrib["whereRight"] = SourceFiles.LineFormat(self.matchNode.xml, True)
            left = myLine
            right = "<--" + self.matchNode.xml.text.replace(' ', nbsp) + "-->"
            self.diffTextToHtml(left, right, node)

    def toText(self):
        return "\n<--" + self.xml.text.replace(' ', nbsp) + "-->\n"

    def cloneTree(self, parent):
        clone = DiffComment(self.xml, parent)
        clone.matchNode = self
        clone.inserted = True
        self.matchNode = clone
        return clone

    def updateCost(self, right):
        if self.xml.text == right.xml.text:
            return 0
        return 3


class DiffElement(DiffRoot):
    def __init__(self, xmlNode, parent):
        if not isinstance(xmlNode, DiffElement):
            DiffRoot.__init__(self, xmlNode, parent)
            preserve = xmlNode.tag == 'artwork'
            if xmlNode.text is not None and xmlNode.text.rstrip() != '':
                n = DiffText(xmlNode.text, xmlNode, self)
                n.preserve = preserve
                self.children.append(n)

            for c in xmlNode.iterchildren():
                n = self.createNode(c, self)
                self.children.append(n)
                if c.tail is not None and c.tail.rstrip() != '':
                    n = DiffText(c.tail, xmlNode, self)
                    n.preserve = preserve
                    self.children.append(n)
                    c.tail = None
            self.FixNamespace(xmlNode.tag)
            if len(self.xml.attrib):
                for key in self.xml.attrib.iterkeys():
                    self.FixNamespace(key)
        else:
            DiffRoot.__init__(self, xmlNode.xml, parent)

    def cloneTree(self, root):
        clone = DiffElement(self, root)
        clone.matchNode = self
        clone.inserted = True
        self.matchNode = clone

        for child in self.children:
            clone.children.append(child.cloneTree(clone))
        return clone

    def clone(self):
        clone = DiffElement(self, None)
        for child in clone.children:
            child.parent = clone
        return clone

    def FixNamespace(self, key):
        global nsKeysIndex
        if key[0] != '{':
            return key
        rbp = key.find('}')
        if rbp >= 0:
            ns = key[1:rbp]
            element = key[rbp+1:]
            if ns not in nsKeys:
                nsNew = 'ns{0}'.format(nsKeysIndex)
                nsKeys[ns] = nsNew
                nsKeysIndex += 1
            return nsKeys[ns] + ':' + element
            # to do this requires returning HTML objects do add in the div elements
            # return '<div class="tooltip"/>' + nsKeys[ns] + '<div class="tooltiptext">' + \
            #     ns + '</div></div>' + ':' + element

    def toText(self):
        text = lxml.etree.tostring(self.xml)
        if type(text).__name__ == 'bytes':
            text = text.decode('utf-8')
        elements = text.split(' ')
        newText = []
        for attr in elements:
            if attr.startswith("xmlns:"):
                continue
            newText.append(attr)
        return ' '.join(newText)

    def ToHtml(self, parent):
        # If we have the right doc info - then emit the <?xml?> line

        root = E.LI()
        parent.append(root)
        anchor = root
        if self.deleted:
            # anchor.attrib['onclick'] = 'return sync2here(1, {0}, -1, 0)'.
            # format(self.xml.sourceline)
            node = E.SPAN()
            node.attrib["class"] = 'left'
            root.attrib["whereLeft"] = SourceFiles.LineFormat(self.xml, False)
            node.text = "<" + self.FixNamespace(self.xml.tag)
            anchor.append(node)
            if len(self.xml.attrib):
                for key in self.xml.attrib.iterkeys():
                    node.text = node.text + " " + self.FixNamespace(key) + \
                                '="' + self.xml.attrib[key] + '"'
        elif self.inserted:
            # anchor.attrib['onclick'] = 'return sync2here(-1, 0, 1, {0})'.
            # format(self.xml.sourceline)
            node = E.SPAN()
            node.attrib['class'] = 'right'
            root.attrib["whereRight"] = SourceFiles.LineFormat(self.xml, True)
            node.text = "<" + self.FixNamespace(self.xml.tag)
            anchor.append(node)
            if len(self.xml.attrib):
                for key in self.xml.attrib.iterkeys():
                    node.text = node.text + " " + self.FixNamespace(key) + \
                                '="' + self.xml.attrib[key] + '"'
        elif self.matchNode is None:
            # anchor.attrib['onclick'] = 'return sync2here(1, {0}, -1, 1)'.
            # format(self.xml.sourceline)
            node = E.SPAN()
            node.attrib['class'] = 'error'
            node.text = "<" + self.FixNamespace(self.xml.tag)
            anchor.append(node)
            if len(self.xml.attrib):
                for key in self.xml.attrib.iterkeys():
                    node.text = node.text + " " + self.FixNamespace(key) + \
                                '="' + self.xml.attrib[key] + '"'
        else:
            root.attrib["whereLeft"] = SourceFiles.LineFormat(self.xml, False)
            root.attrib["whereRight"] = SourceFiles.LineFormat(self.matchNode.xml, True)
            # anchor.attrib['onclick'] = 'return sync2here(1, {0},  1, {1})' \
            #      .format(self.xml.sourceline, self.matchNode.xml.sourceline)
            if self.xml.tag == self.matchNode.xml.tag:
                anchor.text = "<" + self.FixNamespace(self.xml.tag)
            else:
                anchor.text = "<"
                node = E.SPAN()
                node.attrib['class'] = 'left'
                node.text = self.FixNamespace(self.xml.tag)
                anchor.append(node)
                node = E.SPAN()
                node.attrib['class'] = 'right'
                node.text = self.FixNamespace(self.matchNode.xml.tag)
                anchor.append(node)
            if len(self.xml.attrib):
                for key in self.xml.attrib.iterkeys():
                    if key in self.matchNode.xml.attrib:
                        if self.xml.attrib[key] == self.matchNode.xml.attrib[key]:
                            node = E.SPAN()
                            node.text = " " + self.FixNamespace(key) + '="' + \
                                        self.xml.attrib[key] + '"'
                            anchor.append(node)
                        else:
                            leftText = " " + self.FixNamespace(key) + '="' + \
                                       self.xml.attrib[key] + '"'
                            rightText = " " + self.FixNamespace(key) + '="' + \
                                        self.matchNode.xml.attrib[key] + '"'
                            self.diffTextToHtml(leftText, rightText, anchor)
                    else:
                        node = E.SPAN()
                        node.attrib['class'] = 'left'
                        node.text = " " + self.FixNamespace(key) + '="' + \
                                    self.xml.attrib[key] + '"'
                        anchor.append(node)

            for key in self.matchNode.xml.attrib.iterkeys():
                if key not in self.xml.attrib:
                    node = E.SPAN()
                    node.attrib['class'] = 'right'
                    node.text = " " + self.FixNamespace(key) + '="' + \
                                self.matchNode.xml.attrib[key] + '"'
                    anchor.append(node)

        if self.root:
            nsLeft = []
            nsRight = []
            if self.xml.nsmap is not None:
                nsLeft = [self.xml.nsmap[key] for key in self.xml.nsmap]
            if self.matchNode and self.matchNode.xml.nsmap is not None:
                nsRight = [self.matchNode.xml.nsmap[key] for key in self.matchNode.xml.nsmap]

            for ns in nsLeft:
                node = E.SPAN()
                node.text = " xmlns:" + nsKeys[ns] + '="' + ns + '"'
                if ns not in nsRight:
                    node.attrib['class'] = 'left'
                anchor.append(node)

            for ns in nsRight:
                if ns not in nsLeft:
                    node = E.SPAN()
                    node.text = " xmlns:" + nsKeys[ns] + '="' + ns + '"'
                    node.attrib['class'] = 'right'
                    anchor.append(node)

        if len(self.children):
            s = E.SPAN()
            s.text = ">"
            if self.deleted:
                s.attrib['class'] = 'left'
            elif self.inserted:
                s.attrib['class'] = 'right'
            anchor.append(s)
            ul = E.UL()
            for child in self.children:
                child.ToHtml(ul)

            li = E.LI()
            s = E.SPAN()
            li.append(s)
            s.text = "</" + self.FixNamespace(self.xml.tag) + ">"
            if self.deleted:
                s.attrib['class'] = 'left'
                li.attrib["whereLeft"] = SourceFiles.LineFormat(self.xml, False)
            elif self.inserted:
                s.attrib['class'] = 'right'
                li.attrib["whereRight"] = SourceFiles.LineFormat(self.xml, True)
            else:
                li.attrib["whereLeft"] = SourceFiles.LineFormat(self.xml, False)
                li.attrib["whereRight"] = SourceFiles.LineFormat(self.matchNode.xml, True)
                # li.attrib["src"] = "DiffElement-End"

            ul.append(li)
            root.append(ul)
        else:
            s = E.SPAN()
            s.text = "/>"
            if self.deleted:
                s.attrib['class'] = 'left'
            elif self.inserted:
                s.attrib['class'] = 'right'
            anchor.append(s)

    def updateCost(self, right):
        if self.xml.tag == right.xml.tag:
            return 0
        if tagMatching is not None:
            if self.xml.tag in tagMatching and right.xml.tag in tagMatching[self.xml.tag]:
                return 0
        return 100

    def decorateSource(self, sourceLines):
        source = sourceLines[self.xml.sourceLine]
        if self.deleted:
            pass
        elif self.inserted:
            pass
        elif self.matchNode is None:
            pass
        else:
            if self.xml.tag != self.matchNode.xml.tag:
                source.replace("&amp;"+self.xml.tag, "<span class='left'>&amp;" + self.xml.tag)

        sourceLines[self.xml.sourceLine] = source


class DiffText(DiffRoot):
    def __init__(self, text, xmlNode, parent):
        DiffRoot.__init__(self, xmlNode, parent)
        self.text = text

    def cloneTree(self, parent):
        clone = DiffText(self.text, self.xml, parent)
        clone.matchNode = self
        clone.inserted = True
        self.matchNode = clone
        return clone

    def toText(self):
        if not self.preserve:
            return self.text.replace('\n', ' ')
        return self.text

    def ToHtml(self, parent):
        node = E.LI()
        parent.append(node)
        if self.deleted:
            n = E.SPAN()
            n.attrib["class"] = 'left'
            node.attrib["whereLeft"] = SourceFiles.LineFormat(self.xml, False)
            n.text = self.text
            node.append(n)
        elif self.inserted:
            n = E.SPAN()
            n.attrib["class"] = 'right'
            node.attrib["whereRight"] = SourceFiles.LineFormat(self.xml, True)
            n.text = self.text
            node.append(n)
        elif self.matchNode is None:
            n = E.SPAN()
            n.attrib["class"] = 'error'
            n.text = self.text
            node.append(n)
        else:
            node.attrib["whereLeft"] = SourceFiles.LineFormat(self.xml, False)
            node.attrib["whereRight"] = SourceFiles.LineFormat(self.matchNode.xml, True)
            if self.text == self.matchNode.text:
                node.text = self.text
            else:
                self.diffTextToHtml(self.text, self.matchNode.text, node)

    def updateCost(self, right):
        if self.text == right.text:
            return 0
        return 3

    def decorateSource(self, sourceLines):
        pass


class DiffEntity(DiffRoot):
    def __init__(self, node, parent):
        DiffRoot.__init__(self, node, parent)

    def toText(self):
        return self.xml.text

    def ToHtml(self, parent):
        node = E.LI()
        parent.append(node)
        if self.deleted:
            n = E.SPAN()
            n.attrib["class"] = 'left'
            node.attrib['whereLeft'] = SourceFiles.LineFormat(self.xml, False)
            n.text = self.xml.text
            node.append(n)
        elif self.inserted:
            n = E.SPAN()
            n.attrib["class"] = 'right'
            node.attrib['whereRight'] = SourceFiles.LineFormat(self.xml, True)
            n.text = self.xml.text
            node.append(n)
        elif self.matchNode is None:
            n = E.SPAN()
            n.attrib["class"] = 'error'
            n.text = self.xml.text
            node.append(n)
        else:
            node.attrib['whereLeft'] = SourceFiles.LineFormat(self.xml, False)
            node.attrib['whereRight'] = SourceFiles.LineFormat(self.matchNode.xml, True)
            if self.xml.text == self.matchNode.xml.text:
                node.text = self.xml.text
            else:
                self.diffTextToHtml(self.xml.text, self.matchNode.xml.text, node)

    def cloneTree(self, parent):
        clone = DiffEntity(self.xml, parent)
        clone.matchNode = self
        clone.inserted = True
        self.matchNode = clone
        return clone

    def updateCost(self, right):
        if self.xml.text == right.xml.text:
            return 0
        return 100


class DiffParagraph(DiffRoot):
    def __init__(self, node, parent):
        if not isinstance(node, DiffParagraph):
            DiffRoot.__init__(self, node, parent)
        else:
            DiffRoot.__init__(self, node.xml, parent)
        self.preserve = False
        if '{http://www.w3.org/XML/1998/namespace}space' in self.xml.attrib:
            if self.xml.attrib['{http://www.w3.org/XML/1998/namespace}space'] == 'preserve':
                self.preserve = True
        elif self.xml.tag in PreserveSpace:
            self.preserve = True

    def cloneTree(self, root):
        clone = DiffParagraph(self, root)
        clone.matchNode = self
        clone.inserted = True
        self.matchNode = clone

        for child in self.children:
            clone.children.append(child.cloneTree(clone))
        return clone

    def decorateSource(self, sourceLines):
        for child in self.children:
            child.decorateSource(sourceLines)

    def toText(self):
        text = ""
        for child in self.children:
            x = child.toText()
            if type(x).__name__ == 'bytes':
                x = x.decode('utf-8')
            if isinstance(x, DiffComment):
                x = "\n" + text + "\n"
            text += x

        if self.preserve:
            text = text.replace(' ', u'\xa0')
        return text

    def ToHtml(self, parent):

        node = E.LI()
        parent.append(node)
        if self.deleted:
            n = E.SPAN()
            n.attrib["class"] = 'left'
            node.attrib["whereLeft"] = SourceFiles.LineFormat(self.xml, False)
            self.fixPreserveSpace(n, self.toText())
            node.append(n)
        elif self.inserted:
            n = E.SPAN()
            n.attrib['class'] = 'right'
            node.attrib["whereRight"] = SourceFiles.LineFormat(self.xml, True)
            self.fixPreserveSpace(n, self.toText())
            node.append(n)
        elif self.matchNode is None:
            n = E.SPAN()
            n.attrib['class'] = 'error'
            self.fixPreserveSpace(n, self.toText())
            node.append(n)
        else:
            self.diffTextToHtml(self.toText(), self.matchNode.toText(), node)
            node.attrib["whereLeft"] = SourceFiles.LineFormat(self.xml, False)
            node.attrib["whereRight"] = SourceFiles.LineFormat(self.matchNode.xml, True)

    def updateCost(self, right):
        leftText = self.toText()
        rightText = right.toText()
        differ = difflib.SequenceMatcher(a=leftText, b=rightText)
        ratio = differ.quick_ratio()
        return 10 - int(math.floor(ratio*10))



# ----------------------------------------------------
# Copyright The IETF Trust 2018-9, All Rights Reserved
# ----------------------------------------------------

import re
import sys
import six

if six.PY2:
    from aenum import Enum
else:
    from enum import Enum

Console = sys.stdout


class Trace(Enum):
    STOP = 1
    DIAG = 2
    UP = 3
    LEFT = 4


def matrix(left, right):
    if left == right:
        if left[0] == ' ':
            return (0, 2)
        return (0, 1)
    if '\n' in left:
        if '\n' in right:
            return (1, 8)
        return (1, -200)
    elif '\n' in right:
        return (1, -200)
    return (1, -100)


def ComputeEdits(leftArray, rightArray):

    trace = {}
    S = len(leftArray)
    T = len(rightArray)

    # skip over matching at the front
    minLength = min(S, T)
    opStart = ['equal', 0, minLength, 0, minLength]
    rangeStart = 0
    for i in range(minLength):
        if leftArray[i] != rightArray[i]:
            opStart[2] = i
            opStart[4] = i
            rangeStart = i
            break

    if opStart[2] == minLength and minLength == S and minLength == T:
        return [opStart]

    rangeEnd = 0
    opEnd = ['equal', S, S, T, T]
    for i in range(1, minLength-rangeStart-1):
        if leftArray[-i] != rightArray[-i]:
            rangeEnd = i - 1
            opEnd[1] = S-i + 1
            opEnd[3] = T-i + 1
            break

    leftIndex = [i for i in range(rangeStart, len(leftArray)-rangeEnd)]
    rightIndex = [i for i in range(rangeStart, len(rightArray)-rangeEnd)]
    # leftIndex = [ i for i in range(rangeStart, len(leftArray)-rangeEnd)
    #               if leftArray[i] != ' ' ]
    # rightIndex = [ i for i in range(rangeStart, len(rightArray)-rangeEnd)
    #                if rightArray[i] != ' ' ]
    leftIndex.insert(0, 0)
    rightIndex.insert(0, 0)

    S = len(leftIndex)
    T = len(rightIndex)

    gap = 10
    o = gap   # gap
    e = 3     # extend
    maxNegValue = - 655465

    v = {}
    vDiagonal = 0    # best score in cell
    f = maxNegValue  # score from diagonal
    h = maxNegValue  # best score ending with gap from left
    g = {}           # best score ending with gap from above
    g[0] = f
    for i in range(1, T):
        v[i] = -o - (i-1)*e
        g[i] = maxNegValue

    lengthOfHorizontalGap = 0
    lengthOfVerticalGap = {}

    maximumScore = maxNegValue

    trace[0, 0] = Trace.STOP
    for i in range(1, S):
        trace[i, 0] = Trace.UP

    for i in range(1, T):
        trace[0, i] = Trace.LEFT

    #  Fill in the matrices
    for i in range(1, S):   # for all rows
        v[0] = -o - (i - 1) * e

        # Console.write(str(i))
        # Console.write(": \n")
        left = leftArray[leftIndex[i]]

        for j in range(1, T):  # for all columns
            right = rightArray[rightIndex[j]]

            simularityScore = matrix(left, right)

            f = vDiagonal + simularityScore[1]

            # which cell from the left?
            if h - e >= v[j-1] - o:
                h -= e
                lengthOfHorizontalGap += simularityScore[0]
            else:
                h = v[j-1] - o
                lengthOfHorizontalGap = 1

            # which cell from above?
            if g[j] - e >= v[j] - o:
                g[j] = g[j] - e
                lengthOfVerticalGap[j] = lengthOfVerticalGap[j] + simularityScore[0]
            else:
                g[j] = v[j] - o
                lengthOfVerticalGap[j] = 1

            vDiagonal = v[j]
            v[j] = max(f, g[j], h)  # get best score
            if v[j] > maximumScore:
                maximumScore = v[j]

            if v[j] == f:
                trace[i, j] = Trace.DIAG
            elif v[j] == g[j]:
                trace[i, j] = Trace.UP
                # lengths[l] = lengthOfVerticalGap[j]
            else:
                trace[i, j] = Trace.LEFT
                # lengths[l] = lengthOfHorizontalGap

        # PrintLine(i, T, v, g, lengthOfVerticalGap, trace)

        # Reset
        h = maxNegValue
        vDiagonal = 0
        lengthOfHorizontalGap = 0
        # Console.write("\n")

    # Console.write("\n")
    # for i in range(S):
    #    Console.write(str(i))
    #    Console.write(": ")
    #    for j in range(T):
    #        if trace[i,j] == Trace.LEFT:
    #            Console.write('L')
    #        elif trace[i,j] == Trace.UP:
    #            Console.write('U')
    #        elif trace[i,j] == Trace.DIAG:
    #            Console.write('D')
    #        elif trace[i,j] == Trace.STOP:
    #            Console.write('S')
    #        else:
    #            Console.write('X')
    #        Console.write(" ")
    #    Console.write("\n")

    i = S - 1
    j = T - 1

    # val = d[i, j]
    ops = []
    leftIndex.append(leftIndex[-1]+1)
    rightIndex.append(rightIndex[-1]+1)
    ops = []

    op = opEnd
    op2 = None

    stillGoing = True
    while stillGoing:
        if trace[i, j] == Trace.UP:
            op1 = ['remove', leftIndex[i], leftIndex[i+1], rightIndex[j+1], rightIndex[j+1]]
            i -= 1
        elif trace[i, j] == Trace.LEFT:
            op1 = ['insert', leftIndex[i+1], leftIndex[i+1], rightIndex[j], rightIndex[j+1]]
            j -= 1
        elif trace[i, j] == Trace.DIAG:
            if leftArray[leftIndex[i]] == rightArray[rightIndex[j]]:
                op1 = ['equal', leftIndex[i], leftIndex[i+1], rightIndex[j], rightIndex[j+1]]
            else:
                op1 = ['remove', leftIndex[i], leftIndex[i+1], rightIndex[j+1], rightIndex[j+1]]
                op2 = ['insert', leftIndex[i+1], leftIndex[i+1], rightIndex[j], rightIndex[j+1]]
            i -= 1
            j -= 1
        else:
            stillGoing = False
            ops.append(op)
            break

        if op2:
            if op2[0] == op[0]:
                op[1] = op2[1]
                op[3] = op2[3]
            elif op1[0] == op[0]:
                op[1] = op1[1]
                op[3] = op1[3]
                op1 = op2
            else:
                ops.append(op2)
            op2 = None

        if op1[0] == op[0]:
            op[1] = op1[1]
            op[3] = op1[3]
        else:
            ops.append(op)
            op = op1

    # ops.append(op)
    ops.append(opStart)
    ops = list(ops)
    ops.reverse()

    # Console.write("Pre compression\n")
    # for op in ops:
    #     Console.write("%s %2d %2d %2d %2d '%s' '%s'\n" % (op[0], op[1], op[2], op[3], op[4],
    #           ''.join(leftArray[op[1]:op[2]]), ''.join(rightArray[op[3]:op[4]])))
    ops = CompressEdits(ops, leftArray, rightArray)

    return ops


def PrintLine(i, T, v, g, lengthOfVerticalGap, trace):
    Console.write("i={0:3d}".format(i))
    for j in range(1, T):
        Console.write("{0:3d} ".format(v[j]))
    Console.write("\n")

    for j in range(1, T):
        Console.write("{0:3d} ".format(g[j]))
    Console.write("\n")

    for j in range(1, T):
        Console.write("{0:3d} ".format(lengthOfVerticalGap[j]))
    Console.write("\n")

    for j in range(1, T):
        if trace[i, j] == Trace.LEFT:
            ch = 'L'
        elif trace[i, j] == Trace.UP:
            ch = 'U'
        elif trace[i, j] == Trace.DIAG:
            ch = 'D'
        elif trace[i, j] == Trace.STOP:
            ch = 'S'
        else:
            ch = ' '
        Console.write("  {0} ".format(ch))
    Console.write("\n")

    Console.write("\n")


def DoWhiteArray(text):
    result = []
    #  At some point I want to split whitespace with
    #  CR in them to multiple lines
    for right in re.split(r'([\s\xa0=]+)', text, flags=re.UNICODE):
        if len(right) == 0:
            continue
        if right.isspace():
            lastCh = '*'
            for ch in right:
                if ch == ' ':
                    if lastCh != ' ':
                        result.append(ch)
                else:
                    result.append(ch)
                lastCh = ch
        else:
            result.append(right)
    return result


def CompressEdits(ops, leftArray, rightArray):
    opsNew = []
    left = None
    right = None

    for i in range(len(ops)):
        if ops[i][0] == 'remove':
            if leftArray[ops[i][1]][0] == '\n' and right is not None:
                opsNew.append(right)
                right = None
            if left is not None:
                left[2] = ops[i][2]
            else:
                left = ops[i]
        elif ops[i][0] == 'insert':
            if rightArray[ops[i][3]][0] == '\n' and left is not None:
                opsNew.append(left)
                left = None
            if right is not None:
                right[4] = ops[i][4]
            else:
                right = ops[i]
        else:
            if ops[i][2] - ops[i][1] == 1 and \
               (leftArray[ops[i][1]] == u' ' or leftArray[ops[i][1]] == u'\xa0') and \
               i+1 < len(ops) and ops[i+1][0] != 'equal' and not (left is None and right is None):
                if left is not None:
                    left[2] = ops[i][2]
                else:
                    left = ['remove', ops[i][1], ops[i][2], ops[i][3], ops[i][4]]
                if right is not None:
                    right[4] = ops[i][4]
                else:
                    right = ['insert', ops[i][1], ops[i][2], ops[i][3], ops[i][4]]
            else:
                if left is not None:
                    opsNew.append(left)
                    left = None
                if right is not None:
                    opsNew.append(right)
                    right = None
                opsNew.append(ops[i])

    if left is not None:
        opsNew.append(left)
    if right is not None:
        opsNew.append(right)

    return opsNew


if __name__ == '__main__':
    old = " attr1=\"value2\""
    new = " attr1=\"value1\""

    leftArray = DoWhiteArray(old)
    rightArray = DoWhiteArray(new)

    ops = ComputeEdits(leftArray, rightArray)

    for op in ops:
        Console.write("%s %2d %2d %2d %2d '%s' '%s'\n" %
                      (op[0], op[1], op[2], op[3], op[4],
                       ''.join(leftArray[op[1]:op[2]]), ''.join(rightArray[op[3]:op[4]])))

    Console.write("**********************************************\n")

    old = "This is\xa0a\xa0message\nwith one change"
    new = "This is\xa0a message with\ntwo change"
    leftArray = DoWhiteArray(old)
    rightArray = DoWhiteArray(new)

    ops = ComputeEdits(leftArray, rightArray)

    for op in ops:
        Console.write("%s %2d %2d %2d %2d '%s' '%s'\n" %
                      (op[0], op[1], op[2], op[3], op[4],
                       ''.join(leftArray[op[1]:op[2]]), ''.join(rightArray[op[3]:op[4]])))

    Console.write("**********************************************\n")

    old = "This document specifies Elliptic Curve constructs using the curve25519"
    new = "This document specifies elliptic curve constructs using the curve25519"
    leftArray = DoWhiteArray(old)
    rightArray = DoWhiteArray(new)

    ops = ComputeEdits(leftArray, rightArray)

    for op in ops:
        Console.write("%s %2d %2d %2d %2d '%s' '%s'\n" %
                      (op[0], op[1], op[2], op[3], op[4], ''.join(leftArray[op[1]:op[2]]),
                       ''.join(rightArray[op[3]:op[4]])))

    Console.write("**********************************************\n")

    old = "This document specifies algorithm identifiers and ASN.1 encoding formats for " + \
          "Elliptic Curve constructs using the curve25519 and curve448 curves. <!--\xa0" + \
          "Remove\xa0ph\n\xa0The\xa0signature\xa0algorithms\xa0covered\xa0are\xa0Ed25519," + \
          "\xa0Ed25519ph,\xa0Ed448\xa0and\xa0Ed448ph.\n-->\nThe signature algorithms covered" + \
          " are Ed25519 and Ed448. The key agreement algorithm covered are X25519 and X448." + \
          " The encoding for Public Key, Private Key and EdDSA digital signature structures " + \
          "is provided."

    new = "This document specifies algorithm identifiers and ASN.1 encoding formats for " + \
          "elliptic curve constructs using the curve25519 and curve448 curves.  The signature " + \
          "algorithms covered are Ed25519 and Ed448.  The key agreement algorithms covered are " + \
          "X25519 and X448.  The encoding for public key, private key, and Edwards-curve" + \
          " Digital Signature Algorithm (EdDSA) structures is provided."

    leftArray = DoWhiteArray(old)
    rightArray = DoWhiteArray(new)

    ops = ComputeEdits(leftArray, rightArray)

    for op in ops:
        Console.write("%s %2d %2d %2d %2d '%s' '%s'\n" %
                      (op[0], op[1], op[2], op[3], op[4],
                       ''.join(leftArray[op[1]:op[2]]), ''.join(rightArray[op[3]:op[4]])))

    Console.write("**********************************************\n")
    old = "The encoding for Public Key, Private Key and EdDSA digital signature structures " + \
          "is provided."
    new = "The encoding for public key, private key, and Edwards-curve Digital Signature " + \
          "Algorithm (EdDSA) structures is provided."
    leftArray = DoWhiteArray(old)
    rightArray = DoWhiteArray(new)

    ops = ComputeEdits(leftArray, rightArray)

    for op in ops:
        Console.write("%s %2d %2d %2d %2d '%s' '%s'\n" %
                      (op[0], op[1], op[2], op[3], op[4],
                       ''.join(leftArray[op[1]:op[2]]), ''.join(rightArray[op[3]:op[4]])))

    Console.write("**********************************************\n")

    old = "In <xref target=\"RFC8032\"/> the elliptic curve signature system Edwards-curve " + \
          "Digital Signature Algorithm (EdDSA) is described along with a recommendation for " + \
          "the use of the curve25519 and curve448. EdDSA has defined two modes, the PureEdDSA " + \
          "mode without pre-hashing, and the HashEdDSA mode with pre-hashing.\n<!--\xa0Remove" + \
          "\xa0pre-hash\nThe\xa0Ed25519ph\xa0and\xa0Ed448ph\xa0algorithm\xa0definitions\xa0s" + \
          "pecify\xa0the\xa0one-way\xa0hash\xa0function\xa0that\xa0is\xa0used\xa0for\xa0pre-" + \
          "hashing.\nThe\xa0convention\xa0used\xa0for\xa0identifying\xa0the\xa0algorithm/curve" + \
          "\xa0combinations\xa0are\xa0to\xa0use\xa0the\xa0Ed25519\xa0and\xa0Ed448\xa0for\xa0the" + \
          "\xa0PureEdDSA\xa0mode,\xa0with\xa0Ed25519ph\xa0and\xa0Ed448ph\xa0for\xa0the\xa0" + \
          "HashEdDSA\xa0mode.\n-->\nThe convention used for identifying the algorithm/curve " + \
          "combinations is to use \"Ed25519\" and \"Ed448\" for the PureEdDSA mode. The " + \
          "document " + \
          "does not provide the conventions needed for the pre-hash versions of the signature " + \
          "algorithm. The use of the OIDs is described for public keys, private keys and " + \
          "signatures."
    new = "In <xref target=\"RFC8032\"/> the elliptic curve signature system Edwards-curve " + \
          "Digital Signature Algorithm (EdDSA) is described along with a recommendation for " + \
          "the use of the curve25519 and curve448.  EdDSA has defined two modes: the PureEdDSA " + \
          "mode without prehashing and the HashEdDSA mode with prehashing.  The convention " + \
          "used for identifying the algorithm/curve combinations is to use \"Ed25519\" and " + \
          "\"Ed448\" for the PureEdDSA mode.  This document does not provide the conventions " + \
          "needed for the prehash versions of the signature algorithm.  The use of the OIDs is " + \
          "described for public keys, private keys and signatures."

    leftArray = DoWhiteArray(old)
    rightArray = DoWhiteArray(new)
    ops = ComputeEdits(leftArray, rightArray)
    for op in ops:
        Console.write("%s %2d %2d %2d %2d '%s' '%s'\n" %
                      (op[0], op[1], op[2], op[3], op[4],
                       ''.join(leftArray[op[1]:op[2]]), ''.join(rightArray[op[3]:op[4]])))



# ----------------------------------------------------
# Copyright The IETF Trust 2018-9, All Rights Reserved
# ----------------------------------------------------

from xmldiff._zzs import ffi, lib
from xmldiff.DiffNode import DiffParagraph
from xmldiff.EditItem import EditItem

all_handles = []
last_child = None


@ffi.def_extern()
def zzs_get_children(node):
    global last_child
    treeNode = ffi.from_handle(node)
    childs = []
    if not isinstance(treeNode, DiffParagraph):
        for child in treeNode.children:
            h = ffi.new_handle(child)
            childs.append(h)
            all_handles.append(h)

    p = ffi.new("struct cArray *", [len(childs), childs])
    last_child = p
    return p


@ffi.def_extern()
def zzs_insert_cost(node):
    # treeNode = ffi.from_handle(node)
    return 1


@ffi.def_extern()
def zzs_remove_cost(node):
    treeNode = ffi.from_handle(node)
    return treeNode.deleteCost()


@ffi.def_extern()
def zzs_update_cost(node1, node2):
    leftNode = ffi.from_handle(node1)
    rightNode = ffi.from_handle(node2)
    if type(leftNode) is not type(rightNode):
        return 100000
    cost = leftNode.updateCost(rightNode)
    return cost


def distance(leftXml, rightXml, get_children, insertCost, deleteCost, updateCost):
    global all_handles

    ll = ffi.new_handle(leftXml)
    rr = ffi.new_handle(rightXml)
    rv = lib.Distance(ll, rr, lib.zzs_get_children, lib.zzs_insert_cost, lib.zzs_remove_cost,
                      lib.zzs_update_cost)

    editList = []
    for i in range(rv.c):
        editList.append(EditItem(rv.rgEdits[i]))

    all_handles = []

    return editList



# ----------------------------------------------------
# Copyright The IETF Trust 2018-9, All Rights Reserved
# ----------------------------------------------------

from cffi import FFI

ffibuilder = FFI()

ffibuilder.set_source("xmldiff._zzs",
                      r""" // passed to the C compiler
// contains implementation of things declared in cdef()
#include <sys/types.h>

typedef struct eElement {
    int operation;
    void * left;
    void * right;
} eElement;

typedef struct eArray {
  int c;
  eElement rgEdits[];
} eArray;

typedef struct cArray {
  int c;
  void * rg[];
} cArray;

extern struct eArray * Distance(void * leftTree, void *  rightTree,
                                struct cArray  *(*callback)(void *),
                                int (*insert)(void *), int (*remove)(void *),
                                int (*update)(void *, void *));


                      """,
                      sources=["xmldiff/zzs.c"],
                      libraries=[])

ffibuilder.cdef("""

typedef struct eElement {
    int operation;
    void * left;
    void * right;
} eElement;

struct eArray {
  int c;
  eElement rgEdits[];
};

struct cArray {
  int c;
  void * rg[];
};

extern "Python" struct cArray * zzs_get_children(void *);
extern "Python" int zzs_insert_cost(void *);
extern "Python" int zzs_remove_cost(void *);
extern "Python" int zzs_update_cost(void *, void *);


extern struct eArray * Distance(void * leftTree, void *  rightTree,
                                struct cArray  *(*callback)(void *),
                                int (*insert)(void *), int (*remove)(void *),
                                int (*update)(void *, void *));


""")

if __name__ == "__main__":
    ffibuilder.compile(verbose=True)



#!/Usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------
# Copyright The IETF Trust 2018-9, All Rights Reserved
# ----------------------------------------------------
#  Adapted from code by
# Authors: Tim Henderson and Steve Johnson
# Email: tim.tadh@gmail.com, steve@steveasleep.com
# For licensing see the LICENSE file in the top level directory.
#
#  The original code just computed the difference and did not bother with the set of
#  edits that were needed.  We care about the edits themselves.

from __future__ import absolute_import
from six.moves import range

import collections

"""
def py_zeros(dim, pytype):
    assert len(dim) == 2
    return [[pytype for y in range(dim[1])]
            for x in range(dim[0])]
"""

try:
    from editdist import distance as strdist
except ImportError:
    def strdist(a, b):
        if a == b:
            return 0
        else:
            return 1


class EditItem(object):
    __slots__ = ['operation', 'left', 'right', 'cost']
    OP_INSERT = 1
    OP_DELETE = 2
    OP_RENAME = 3
    OP_MATCH = 4
    OP_COMBINE = 5

    def __init__(self):
        self.cost = 0
        self.operation = 0
        self.left = None
        self.right = None
        pass

    def reset(self):
        self.operation = 0
        self.cost = 0
        self.left = None
        self.right = None

    def setOperation(self, operation, left, right):
        self.operation = operation
        self.left = left
        self.right = right
        self.cost = 0
        if left is not None and isinstance(left, EditItem):
            self.cost += left.cost
        if right is not None and isinstance(right, EditItem):
            self.cost += right.cost

    def clone(self):
        cloneOut = EditItem()
        returnValue = cloneOut
        cloneOut.operation = self.operation
        cloneOut.cost = self.cost
        stack = []
        if self.left is not None:
            stack.append((1, cloneOut, self))
        if self.right is not None:
            stack.append((0, cloneOut, self))

        while len(stack) > 0:
            side, node, node1 = stack.pop()
            xx = node1.left if side else node1.right

            if xx is not None:
                if isinstance(xx, EditItem):
                    if xx.operation == 0:
                        continue
                    if xx.operation == EditItem.OP_COMBINE:
                        newNode = EditItem()
                        newNode.operation = xx.operation
                        newNode.cost = xx.cost
                        if xx.left is not None:
                            stack.append((1, newNode, xx))
                        if xx.right is not None:
                            stack.append((0, newNode, xx))
                    else:
                        newNode = xx
                else:
                    newNode = xx
                if side:
                    node.left = newNode
                else:
                    node.right = newNode

        return returnValue

    def toList(self):
        list = []
        self._toList(list)
        return list

    def _toList(self, list):
        if self.operation != EditItem.OP_COMBINE:
            list.append(self)
            return
        if self.left is not None and isinstance(self.left, EditItem):
            self.left._toList(list)
        if self.right is not None and isinstance(self.right, EditItem):
            self.right._toList(list)

    def toString(self):
        left = ""
        right = ""
        if self.left:
            left = "left= {:>3}".format(self.left.index)
        if self.right:
            right = " right= {:>3}".format(self.right.index)
        txt = "OP={0} {1:>9}{2}".format(str(self.operation), left, right)
        return txt


class AnnotatedTree(object):

    def __init__(self, root, get_children):
        self.get_children = get_children

        self.root = root
        self.nodes = list()  # a post-order enumeration of the nodes in the tree
        self.ids = list()    # a matching list of ids
        self.lmds = list()   # left most descendents
        self.keyroots = None

        # k and k' are nodes specified in the post-order enumeration.
        # keyroots = {k | there exists no k'>k such that lmd(k) == lmd(k')}
        # see paper for more on keyroots

        stack = list()
        pstack = list()
        stack.append((root, collections.deque()))
        j = 0
        while len(stack) > 0:
            n, anc = stack.pop()
            nid = j
            for c in self.get_children(n):
                a = collections.deque(anc)
                a.appendleft(nid)
                stack.append((c, a))
            pstack.append(((n, nid), anc))
            j += 1
        lmds = dict()
        keyroots = dict()
        i = 0
        while len(pstack) > 0:
            (n, nid), anc = pstack.pop()
            # print list(anc)
            self.nodes.append(n)
            self.ids.append(nid)
            # print n.label, [a.label for a in anc]
            if not self.get_children(n):
                lmd = i
                for a in anc:
                    if a not in lmds:
                        lmds[a] = i
                    else:
                        break
            else:
                try:
                    lmd = lmds[nid]
                except Exception:
                    import pdb
                    pdb.set_trace()
            self.lmds.append(lmd)
            keyroots[lmd] = i
            i += 1
        self.keyroots = sorted(keyroots.values())


def distance(A, B, get_children, insert_cost, remove_cost, update_cost):
    '''Computes the exact tree edit distance between trees A and B with a
    richer API than :py:func:`zss.simple_distance`.
    Use this function if either of these things are true:
    * The cost to insert a node is **not** equivalent to the cost of changing
      an empty node to have the new node's label
    * The cost to remove a node is **not** equivalent to the cost of changing
      it to a node with an empty label
    Otherwise, use :py:func:`zss.simple_distance`.
    :param A: The root of a tree.
    :param B: The root of a tree.
    :param get_children:
        A function ``get_children(node) == [node children]``.  Defaults to
        :py:func:`zss.Node.get_children`.
    :param insert_cost:
        A function ``insert_cost(node) == cost to insert node >= 0``.
    :param remove_cost:
        A function ``remove_cost(node) == cost to remove node >= 0``.
    :param update_cost:
        A function ``update_cost(a, b) == cost to change a into b >= 0``.
    :return: An integer distance [0, inf+)
    '''
    A, B = AnnotatedTree(A, get_children), AnnotatedTree(B, get_children)
    # treedists = zeros((len(A.nodes), len(B.nodes)), int)
    treedists = [[None for x in range(len(B.nodes))] for y in range(len(A.nodes))]
    fd = [[EditItem() for x in range(len(B.nodes)+2)] for y in range(len(A.nodes)+2)]

    A_remove = [remove_cost(A.nodes[i]) for i in range(len(A.nodes))]
    B_insert = [insert_cost(B.nodes[i]) for i in range(len(B.nodes))]

    def treedist(i, j):
        Al = A.lmds
        Bl = B.lmds
        An = A.nodes
        Bn = B.nodes

        m = i - Al[i] + 2
        n = j - Bl[j] + 2

        for x in range(1, m):
            fd1 = fd[x]
            for y in range(1, n):
                fd1[y].reset()

        ioff = Al[i] - 1
        joff = Bl[j] - 1

        for x in range(1, m):  # δ(l(i1)..i, θ) = δ(l(1i)..1-1, θ) + γ(v → λ)
            fd[x][0].setOperation(EditItem.OP_COMBINE, fd[x-1][0], A_remove[x+ioff])
        fd1 = fd[0]
        for y in range(1, n):  # δ(θ, l(j1)..j) = δ(θ, l(j1)..j-1) + γ(λ → w)
            fd1[y].setOperation(EditItem.OP_COMBINE, fd1[y-1], B_insert[y+joff])

        for x in range(1, m):  # the plus one is for the xrange impl
            fd1 = fd[x]
            fdm1 = fd[x-1]
            x_ioff = x + ioff
            t = Al[i] == Al[x_ioff]
            remove = A_remove[x_ioff]
            treedists_1 = treedists[x_ioff]
            left = An[x_ioff]
            p = Al[x_ioff]-1-ioff
            fd_p = fd[p]
            y_joff = joff
            for y in range(1, n):
                y_joff += 1
                # only need to check if x is an ancestor of i
                # and y is an ancestor of j
                if t and Bl[j] == Bl[y_joff]:
                    #                   +-
                    #                   | δ(l(i1)..i-1, l(j1)..j) + γ(v → λ)
                    # δ(F1 , F2 ) = min-+ δ(l(i1)..i , l(j1)..j-1) + γ(λ → w)
                    #                   | δ(l(i1)..i-1, l(j1)..j-1) + γ(v → w)
                    #                   +-

                    insert = B_insert[y_joff]
                    update = update_cost(left, Bn[y_joff])

                    op1Cost = fdm1[y].cost + remove.cost
                    op2Cost = fd1[y-1].cost + insert.cost
                    op3Cost = fdm1[y-1].cost + update.cost

                    if op1Cost < op2Cost:
                        if op1Cost < op3Cost:
                            fd1[y].setOperation(EditItem.OP_COMBINE, fdm1[y], remove)
                        elif op2Cost < op3Cost:
                            fd1[y].setOperation(EditItem.OP_COMBINE, fd1[y-1], insert)
                        else:
                            fd1[y].setOperation(EditItem.OP_COMBINE, fdm1[y-1], update)
                    else:
                        if op2Cost < op3Cost:
                            fd1[y].setOperation(EditItem.OP_COMBINE, fd1[y-1], insert)
                        else:
                            fd1[y].setOperation(EditItem.OP_COMBINE, fdm1[y-1], update)

                    # fd[x][y] = min(
                    #     fd[x-1][y] + remove_cost(An[x+ioff]),
                    #     fd[x][y-1] + insert_cost(Bn[y+joff]),
                    #     fd[x-1][y-1] + update_cost(An[x+ioff], Bn[y+joff]),
                    # )
                    treedists_1[y_joff] = fd1[y].clone()
                else:
                    #                   +-
                    #                   | δ(l(i1)..i-1, l(j1)..j) + γ(v → λ)
                    # δ(F1 , F2 ) = min-+ δ(l(i1)..i , l(j1)..j-1) + γ(λ → w)
                    #                   | δ(l(i1)..l(i)-1, l(j1)..l(j)-1)
                    #                   |                     + treedist(i1,j1)
                    #                   +-
                    q = Bl[y_joff]-1-joff

                    insert = B_insert[y_joff]

                    op1Cost = fdm1[y].cost + remove.cost
                    op2Cost = fd1[y-1].cost + insert.cost
                    op3Cost = fd_p[q].cost + treedists_1[y_joff].cost

                    if op1Cost < op2Cost:
                        if op1Cost < op3Cost:
                            fd1[y].setOperation(EditItem.OP_COMBINE, fdm1[y], remove)
                        elif op2Cost < op3Cost:
                            fd1[y].setOperation(EditItem.OP_COMBINE, fd1[y-1], insert)
                        else:
                            fd1[y].setOperation(EditItem.OP_COMBINE, fd_p[q], treedists_1[y_joff])
                    else:
                        if op2Cost < op3Cost:
                            fd1[y].setOperation(EditItem.OP_COMBINE, fd1[y-1], insert)
                        else:
                            fd1[y].setOperation(EditItem.OP_COMBINE, fd_p[q], treedists_1[y_joff])

                    # fd[x][y] = min(
                    #     fd[x-1][y] + remove_cost(An[x+ioff]),
                    #     fd[x][y-1] + insert_cost(Bn[y+joff]),
                    #     fd[p][q] + treedists[x+ioff][y+joff]
                    # )

    for i in A.keyroots:
        for j in B.keyroots:
            treedist(i, j)

    return treedists[-1][-1]



# ----------------------------------------------------
# Copyright The IETF Trust 2018-9, All Rights Reserved
# ----------------------------------------------------

from xmldiff._zzs import ffi


class EditItem(object):
    __slots__ = ['operation', 'left', 'right', 'cost']
    OP_INSERT = 1
    OP_DELETE = 2
    OP_RENAME = 3
    OP_MATCH = 4
    OP_COMBINE = 5

    def __init__(self, data):
        self.cost = 0
        self.operation = data.operation
        if data.left is not None and data.left != ffi.NULL:
            self.left = ffi.from_handle(data.left)
        else:
            self.left = None
        self.right = None
        if data.right is not None and data.right != ffi.NULL:
            self.right = ffi.from_handle(data.right)

    def toString(self):
        left = ""
        right = ""
        if self.left:
            left = "left= {:>3}".format(self.left.index)
        if self.right:
            right = " right= {:>3}".format(self.right.index)
        txt = "OP={0} {1:>9}{2}".format(str(self.operation), left, right)
        return txt
